/* Copyright (C) 2007-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Igor Kulakov, Sergey Gorbunov [committer], Maksym Zyzak */

/// \file CaHit.h
/// \brief ca::Hit class describes a generic hit for the CA tracker
/// \date 2007-2023

#ifndef CaHit_h
#define CaHit_h 1

#include <boost/serialization/access.hpp>

#include "CaSimd.h"

namespace cbm::algo::ca
{
  using HitIndex_t   = unsigned int;  ///< Index of ca::Hit
  using StripIndex_t = unsigned int;  ///< Index of the station strip


  /// \brief ca::Hit class describes a generic hit for the CA tracker
  ///
  struct Hit {
    friend class boost::serialization::access;

    /// NOTE: For STS f and b correspond to the indexes of the front and back clusters of the hit in a dataset. For other
    ///       tracking detectors (MVD, MuCh, TRD, TOF) f == b and corresponds to the index of the hit. Indexes f and b
    ///       do not intersect between different detector stations.
    StripIndex_t f {0};  ///< front hit key index
    StripIndex_t b {0};  ///< back hit key index

    fscal x {0.};       ///< measured X coordinate [cm]
    fscal y {0.};       ///< measured Y coordinate [cm]
    fscal z {0.};       ///< fixed Z coordinate [cm]
    fscal t {0.};       ///< measured time [ns]
    fscal dx2 {0.};     ///< rms^2 of uncertainty of X coordinate [cm]
    fscal dy2 {0.};     ///< rms^2 of uncertainty of Y coordinate [cm]
    fscal dxy {0.};     ///< X/Y covariance [cm2]
    fscal dt2 {0.};     ///< measured uncertainty of time [ns]
    fscal rangeX {0.};  ///< +/- range of uncertainty of X coordinate [cm]
    fscal rangeY {0.};  ///< +/- range of uncertainty of Y coordinate [cm]
    fscal rangeT {0.};  ///< +/- range of uncertainty of time [ns]

    HitIndex_t Id {0};  ///< id of the hit
    int iSt {-1};       ///< index of station in the active stations array

  private:
    /// Serialization method, used to save ca::Hit objects into binary or text file in a defined order
    template<class Archive>
    void serialize(Archive& ar, const unsigned int /*version*/)
    {
      ar& f;
      ar& b;
      ar& x;
      ar& y;
      ar& z;
      ar& t;
      ar& dx2;
      ar& dxy;
      ar& dy2;
      ar& dt2;
      ar& rangeX;
      ar& rangeY;
      ar& rangeT;
      ar& Id;
      ar& iSt;
    }
  };

}  // namespace cbm::algo::ca

#endif
