/* Copyright (C) 2007-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer], Igor Kulakov, Maksym Zyzak, Sergei Zharko */

#ifndef CaField_h
#define CaField_h 1

#include <boost/serialization/access.hpp>

#include <functional>
#include <string>

#include "CaConstants.h"
#include "CaSimd.h"
#include "CaTrackParam.h"

namespace cbm::algo::ca
{

  class FieldValue {

  public:
    fvec x {0.f};  //< x-component of the field
    fvec y {0.f};  //< y-component of the field
    fvec z {0.f};  //< z-component of the field

    /// Combines the magnetic field with another field value object using weight
    /// \param B  other field value to combine with
    /// \param w  weight from 0 to 1
    void Combine(const FieldValue& B, const fvec& w);

    void Combine(const FieldValue& B, const fmask& w);

    /// Consistency checker
    void CheckConsistency() const;

    /// Operator << overloading
    friend std::ostream& operator<<(std::ostream& out, const FieldValue& B);

    /// String representation of class contents
    /// \param indentLevel      number of indent characters in the output
    std::string ToString(int indentLevel) const;

    /// Serialization function
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      ar& x;
      ar& y;
      ar& z;
    }
  } _fvecalignment;

  [[gnu::always_inline]] inline void FieldValue::Combine(const FieldValue& B, const fvec& w)
  {
    x += w * (B.x - x);
    y += w * (B.y - y);
    z += w * (B.z - z);
  }

  [[gnu::always_inline]] inline void FieldValue::Combine(const FieldValue& B, const fmask& w)
  {
    x(w) = B.x;
    y(w) = B.y;
    z(w) = B.z;
  }

  /// Class represents a set of magnetic field approximation coefficients
  ///
  // TODO: Crosscheck the default content (S.Zharko)
  class FieldSlice {
  public:
    /// Default constructor
    FieldSlice();

    /// Consistency checker
    void CheckConsistency() const;

    /// Gets field value from (x, y) fvec point
    /// \param x  x-coordinate of input
    /// \param y  y-coordinate of input
    /// \return B  the FieldValue output
    FieldValue GetFieldValue(const fvec& x, const fvec& y) const;

    /// Gets field value for the intersection with a straight track
    /// \param t  straight track
    /// \return B  the FieldValue output
    FieldValue GetFieldValueForLine(const TrackParamV& t) const;

    /// String representation of class contents
    /// \param indentLevel      number of indent characters in the output
    std::string ToString(int indentLevel = 0) const;

  public:
    // NOTE: We don't use an initialization of arrays here because we cannot be sure
    //       if the underlying type (fvec) has a default constructor, but
    //       we are sure, that it can be initialized with a float. (S.Zharko)
    static constexpr auto kMaxNFieldApproxCoefficients = constants::size::MaxNFieldApproxCoefficients;
    fvec cx[kMaxNFieldApproxCoefficients];  ///< Polynomial coefficients for x-component of the field value
    fvec cy[kMaxNFieldApproxCoefficients];  ///< Polynomial coefficients for y-component of the field value
    fvec cz[kMaxNFieldApproxCoefficients];  ///< Polynomial coefficients for z-component of the field value

    fvec z {constants::Undef<fscal>};  ///< z coordinate of the slice

    /// Serialization function
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      ar& cx;
      ar& cy;
      ar& cz;
      ar& z;
    }
  } _fvecalignment;


  class FieldRegion {
  public:
    // NOTE: When a custom constructor is defined, default constructor also should be provided (S.Zharko)
    FieldRegion() = default;

    FieldRegion(float reg[10]) noexcept;

    /// Consistency checker
    void CheckConsistency() const;

    /// Gets the field vector and writes it into B pointer
    /// \param x  x-coordinate of the point to calculate the field
    /// \param y  y-coordinate of the point to calculate the field
    /// \param z  z-coordinate of the point to calculate the field
    /// \return    the magnetic field value
    FieldValue Get(const fvec& x, const fvec& y, const fvec& z) const;

    /// Interpolates the magnetic field by three nodal points and sets the result to this FieldRegion object
    /// The result is a quadratic interpolation of the field as a function of z
    /// \param b0   field value in the first nodal point
    /// \param b0z  z-coordinate of the first nodal point
    /// \param b1   field value in the second nodal point
    /// \param b1z  z-coordinate of the second nodal point
    /// \param b2   field value in the third nodal point
    /// \param b2z  z-coordinate of the third nodal point
    /// TODO: does the sequence of b0z, b1z and b2z matter? If yes, probalby we need an assertion (S.Zharko)
    void Set(const FieldValue& b0, const fvec b0z, const FieldValue& b1, const fvec b1z, const FieldValue& b2,
             const fvec b2z);

    /// Interpolates the magnetic field by thwo nodal points and sets the result to this FieldRegion object.
    /// The result is a linear interpolation of the field as a function of z
    /// \param b0   field value in the first nodal point
    /// \param b0z  z-coordinate of the first nodal point
    /// \param b1   field value in the second nodal point
    /// \param b1z  z-coordinate of the second nodal point
    /// TODO: does the sequence of b0z and b1z matter? If yes, probalby we need an assertion (S.Zharko)
    void Set(const FieldValue& b0, const fvec b0z, const FieldValue& b1, const fvec b1z);

    /// Shifts the coefficients to new central point
    /// \param z  z-coordinate of the new central point
    void Shift(fvec z);

    /// Replaces the selected layer of the coefficients with one from another
    /// FieldRegion object
    /// \param i0    the index of the fvect layer in this FieldRegion object
    /// \param fl    the other FieldRegion object, which layer is going to be replaced
    /// \param i1    the index of the source fvect layer to copy
    void SetOneEntry(const int i0, const FieldRegion& f1, const int i1);

    /// Replaces all the layers of the coefficients with one selected layer from another
    /// FieldRegion object
    /// \param fl    the other FieldRegion object, which layer is going to be replaced
    /// \param i1    the index of the source fvect layer to copy
    void SetOneEntry(const FieldRegion& f1, const int i1);

    /// String representation of class contents
    /// \param indentLevel      number of indent characters in the output
    std::string ToString(int indentLevel = 0) const;

    /// Use original field instead of the field approximation
    void SetUseOriginalField(bool v = true) { fUseOriginalField = v; }

    /// Force using the original field
    static void ForceUseOfOriginalField(bool v = true) { fgForceUseOfOriginalField = v; }

    /// Set the original field
    static void SetOriginalField(std::function<std::tuple<double, double, double>(double x, double y, double z)> f)
    {
      fgOdiginalField = f;
    }

  public:
    static bool fgForceUseOfOriginalField;
    static std::function<std::tuple<double, double, double>(double x, double y, double z)> fgOdiginalField;

    // TODO: Probably it's better to have arrays instead of separate fvec values? (S.Zharko)
    // Bx(z) = cx0 + cx1*(z-z0) + cx2*(z-z0)^2
    fvec cx0 {0.f};
    fvec cx1 {0.f};
    fvec cx2 {0.f};

    // By(z) = cy0 + cy1*(z-z0) + cy2*(z-z0)^2
    fvec cy0 {0.f};
    fvec cy1 {0.f};
    fvec cy2 {0.f};

    // Bz(z) = cz0 + cz1*(z-z0) + cz2*(z-z0)^2
    fvec cz0 {0.f};
    fvec cz1 {0.f};
    fvec cz2 {0.f};

    fvec z0 {0.f};  ///< z-coordinate of the field region central point

    bool fUseOriginalField {false};

    /// Serialization function
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      ar& cx0;
      ar& cx1;
      ar& cx2;
      ar& cy0;
      ar& cy1;
      ar& cy2;
      ar& cz0;
      ar& cz1;
      ar& cz2;
    }
  } _fvecalignment;

}  // namespace cbm::algo::ca
#endif
