/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CaMonitor.h
/// @brief  CA Tracking monitor class
/// @since  25.08.2023
/// @author S.Zharko <s.zharko@gsi.de>

#ifndef CaMonitor_h
#define CaMonitor_h 1

#include <iomanip>
#include <sstream>
#include <string>

#include "CaEnumArray.h"

namespace cbm::algo::ca
{
  /// @class  Monitor
  /// @brief  Monitor class for CA tracking
  /// @tparam EMonitorKey  A enum class, containing keys for monitorables
  ///
  template<class EMonitorKey>
  class Monitor {
  public:
    /// @brief Alias to array, indexed by Monitorable enumeration
    template<typename T>
    using MonitorableArr_t = EnumArray<EMonitorKey, T>;

    /// @brief Constructor
    /// @param name Name of monitor
    Monitor(const std::string& name) : fsName(name) {}

    /// @brief Gets key name
    /// @param key Monitorable key
    const std::string& GetKeyName(EMonitorKey key) const { return faKeyNames[key]; }

    /// @brief Gets counter value
    /// @param key
    int GetValue(EMonitorKey key) const { return faKeyCounters[key]; }

    /// @brief Increments key counter by 1
    /// @param key Monitorable key
    void Increment(EMonitorKey key) { ++faKeyCounters[key]; };

    /// @brief Increments key counter by a number
    /// @param key  Monitorable key
    /// @param num  Number to add
    void Increment(EMonitorKey key, int num) { faKeyCounters[key] += num; }

    /// @brief Resets the counters
    void Reset() { faKeyCounters.fill(0); }

    /// @brief Sets keys of monitorables, which are used as denominators for ratios
    /// @param vKeys Vector of keys
    void SetRatioKeys(const std::vector<EMonitorKey>& vKeys) { fvRatioKeys = vKeys; }

    /// @brief Sets name of key
    /// @param name  Name of monitoralble
    void SetKeyName(EMonitorKey key, const char* name) { faKeyNames[key] = name; }

    /// @brief Prints counters summary to string
    std::string ToString() const;

  private:
    std::vector<EMonitorKey> fvRatioKeys {};      ///< List of keys, which are used as denominators in ratios
    std::string fsName;                           ///< Name of the monitor
    MonitorableArr_t<std::string> faKeyNames {};  ///< Names of keys
    MonitorableArr_t<int> faKeyCounters {};       ///< Counters of keys
  };

  // *****************************************
  // **  Template function implementations  **
  // *****************************************


  // ---------------------------------------------------------------------------------------------------------------------
  //
  template<class EMonitorKey>
  std::string Monitor<EMonitorKey>::ToString() const
  {
    using std::left;
    using std::right;
    using std::setfill;
    using std::setw;
    std::stringstream msg;
    constexpr size_t width = 24;
    size_t fillWidth       = (width + 2) * (1 + fvRatioKeys.size());
    msg << "----- Monitor: " << fsName << ' ' << setw(fillWidth - fsName.size() - 16) << setfill('-') << '-' << '\n';
    msg << setfill(' ');
    msg << setw(width) << left << "Key" << ' ';
    msg << setw(width) << left << "Total" << ' ';
    for (auto key : fvRatioKeys) {
      msg << setw(width) << left << std::string("per ") + faKeyNames[key] << ' ';
    }
    msg << '\n';
    for (int iKey = 0; iKey < static_cast<int>(EMonitorKey::kEND); ++iKey) {
      msg << setw(width) << left << faKeyNames[iKey] << ' ';
      msg << setw(width) << right << faKeyCounters[iKey] << ' ';
      for (auto keyDen : fvRatioKeys) {
        auto ratio = static_cast<double>(faKeyCounters[iKey]) / faKeyCounters[keyDen];
        msg << setw(width) << right << ratio << ' ';
      }
      msg << '\n';
    }
    msg << setw(fillWidth) << setfill('-') << '-' << '\n';
    msg << setfill(' ');
    return msg.str();
  }
}  // namespace cbm::algo::ca

#endif  // CaMonitor_h
