/* Copyright (C) 2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */

#include "EventBuilder.h"

#include <cassert>
#include <iomanip>

using std::is_sorted;
using std::vector;

namespace cbm::algo::evbuild
{

  // -----   Algorithm execution   --------------------------------------------
  EventBuilder::resultType EventBuilder::operator()(const DigiData& ts, const vector<double> triggers) const
  {
    // --- Output data
    resultType result = {};
    result.first.resize(triggers.size());

    std::transform(triggers.begin(), triggers.end(), result.first.begin(),
                   [&ts, &result, this](const double& trigger) { return BuildEvent(ts, result.second, trigger); });

    EventBuilderMonitorData& monitor = result.second;
    monitor.fSts.fNum += ts.fSts.size();
    monitor.fRich.fNum += ts.fRich.size();
    monitor.fMuch.fNum += ts.fMuch.size();
    monitor.fTrd.fNum += ts.fTrd.size();
    monitor.fTrd2d.fNum += ts.fTrd2d.size();
    monitor.fTof.fNum += ts.fTof.size();
    monitor.fPsd.fNum += ts.fPsd.size();
    monitor.fFsd.fNum += ts.fFsd.size();
    monitor.fBmon.fNum += ts.fBmon.size();
    return result;
  }

  // --- Build a single event
  DigiEvent EventBuilder::BuildEvent(const DigiData& ts, EventBuilderMonitorData& monitor, double trigger) const
  {
    DigiEvent event;
    event.fTime = trigger;

    // --- Loop over systems
    for (auto entry : fConfig.fWindows) {

      auto system       = entry.first;
      const double tMin = trigger + entry.second.first;
      const double tMax = trigger + entry.second.second;

      // --- Build the event using trigger window
      switch (system) {
        case ECbmModuleId::kSts: {
          event.fSts = CopyRange(ts.fSts, tMin, tMax);
          break;
        }
        case ECbmModuleId::kRich: {
          event.fRich = CopyRange(ts.fRich, tMin, tMax);
          break;
        }
        case ECbmModuleId::kMuch: {
          event.fMuch = CopyRange(ts.fMuch, tMin, tMax);
          break;
        }
        case ECbmModuleId::kTrd: {
          event.fTrd = CopyRange(ts.fTrd, tMin, tMax);
          break;
        }
        case ECbmModuleId::kTrd2d: {
          event.fTrd2d = CopyRange(ts.fTrd2d, tMin, tMax);
          break;
        }
        case ECbmModuleId::kTof: {
          event.fTof = CopyRange(ts.fTof, tMin, tMax);
          break;
        }
        case ECbmModuleId::kPsd: {
          event.fPsd = CopyRange(ts.fPsd, tMin, tMax);
          break;
        }
        case ECbmModuleId::kFsd: {
          event.fFsd = CopyRange(ts.fFsd, tMin, tMax);
          break;
        }
        case ECbmModuleId::kT0: {
          event.fBmon = CopyRange(ts.fBmon, tMin, tMax);
          break;
        }
        default: break;
      }
    }
    monitor.fSts.fNumInEvents += event.fSts.size();
    monitor.fRich.fNumInEvents += event.fRich.size();
    monitor.fMuch.fNumInEvents += event.fMuch.size();
    monitor.fTrd.fNumInEvents += event.fTrd.size();
    monitor.fTrd2d.fNumInEvents += event.fTrd2d.size();
    monitor.fTof.fNumInEvents += event.fTof.size();
    monitor.fPsd.fNumInEvents += event.fPsd.size();
    monitor.fFsd.fNumInEvents += event.fFsd.size();
    monitor.fBmon.fNumInEvents += event.fBmon.size();
    return event;
  }
  // --------------------------------------------------------------------------


  // -----   Info to string   -------------------------------------------------
  std::string EventBuilder::ToString() const
  {
    std::stringstream out;
    out << "--- Using EventBuilder with event windows:";
    for (const auto& entry : fConfig.fWindows) {
      out << "\n  " << std::left << std::setw(5) << ::ToString(entry.first) << ": ";
      out << "  [" << std::right << std::setw(5) << entry.second.first;
      out << ", " << std::right << std::setw(5) << entry.second.second << "] ns";
    }
    return out.str();
  }
  // --------------------------------------------------------------------------

}  // namespace cbm::algo::evbuild
