/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/***************************************************************************************************
 * @file   CbmTrackingDetectorInterfaceBase.h
 * @brief  Base abstract class for tracking detector interface to L1
 * @since  31.05.2022
 * @author S.Zharko <s.zharko@gsi.de>
 ***************************************************************************************************/

#ifndef CbmTrackingDetectorInterfaceBase_h
#define CbmTrackingDetectorInterfaceBase_h 1

#include "CbmHit.h"
#include "CbmPixelHit.h"

#include "FairMCPoint.h"

#include <array>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include <cmath>

/// @class CbmTrackingDetectorInterfaceBase
/// @brief Abstract class, which should be inherited by every detecting subsystem tracking interface class
///
class CbmTrackingDetectorInterfaceBase {
public:
  /// @brief Virtual destructor
  virtual ~CbmTrackingDetectorInterfaceBase() {}

  /// @brief Checks detector interface: boundary conditions of the parameters
  bool Check() const;

  /// @brief Returns the name of the detector subsystem
  virtual std::string GetDetectorName() const = 0;

  /// @brief Gets actual number of stations, provided by the current geometry setup
  virtual int GetNtrackingStations() const = 0;

  /// @brief Gets stereo angles of the two independent measured coordinates
  /// @note  The tracking does not use this method. It is only used by the QA task.
  /// @param  address  detector unique identifier
  /// @return [phiU, phiV] - Stereo angles [rad]
  virtual std::tuple<double, double> GetStereoAnglesSensor(int address) const = 0;

  /// @brief  Gets a tracking station of a CbmHit
  /// @param  hit  A pointer to CbmHit
  /// @return Local index of the tracking station
  virtual int GetTrackingStationIndex(const CbmHit* hit) const = 0;

  /// @brief  Gets a tracking station of a FairMCPoint
  /// @param  point  A pointer to FairMCHit
  /// @return Local index of the tracking station
  virtual int GetTrackingStationIndex(const FairMCPoint* point) const = 0;

  /// @brief  Gets a tracking station by the address of element
  /// @param  address  Unique element address
  /// @return Local index of the tracking station
  virtual int GetTrackingStationIndex(int address) const = 0;

  /// @brief  Gets upper bound of a station along the X-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the X-axis [cm]
  virtual double GetXmax(int stationId) const = 0;

  /// @brief  Gets lower bound of a station along the X-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the X-axis [cm]
  virtual double GetXmin(int stationId) const = 0;

  /// @brief  Gets upper bound of a station along the Y-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the Y-axis [cm]
  virtual double GetYmax(int stationId) const = 0;

  /// @brief  Gets lower bound of a station along the Y-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the Y-axis [cm]
  virtual double GetYmin(int stationId) const = 0;

  /// @brief  Gets reference z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Reference z position of the station [cm]
  virtual double GetZref(int stationId) const = 0;

  /// @brief  Gets reference z of the detector module (e.g., RPC for TOF)
  /// @param  address  Address of the module
  /// @return Reference z of module (if not defined, the ref z of the station will be returned)
  virtual double GetZrefModule(int address) { return GetZref(GetTrackingStationIndex(address)); }

  /// @brief  Gets min z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return min Z of the station [cm]
  virtual double GetZmin(int stationId) const = 0;

  /// @brief  Gets max z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return max Z of the station [cm]
  virtual double GetZmax(int stationId) const = 0;

  /// @brief  Check if station provides time measurements
  /// @param  stationId  Tracking station ID in the setup
  /// @return Flag: true - station provides time measurements, false - station does not provide time measurements
  virtual bool IsTimeInfoProvided(int stationId) const = 0;

  /// @brief  Gets x,y,t ranges of a CbmPixelHit
  /// @param  hit  A hit
  /// @return range X, Y, T
  std::tuple<double, double, double> GetHitRanges(const CbmPixelHit& hit) const
  {
    // by default assume gaussian distributions of errors
    return std::tuple(3.5 * hit.GetDx(), 3.5 * hit.GetDy(), 3.5 * hit.GetTimeError());
  }

  /// Technical flag: true - all downcasts are done with dynamic_cast followed by checks for nullptr (increases
  /// computing time, better for debug), false - all downcasts are done with static_cast without sanity checks
  /// (decreases computing time, but can cause bugs)
  static constexpr bool kUseDynamicCast {true};

  /// @brief Prints all the parameters into table and saves the table as a string
  std::string ToString() const;
};

#endif  // CbmTrackingDetectorInterfaceBase_h
