/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/***************************************************************************************************
 * @file   CbmTrdTrackingInterface.h
 * @brief  Input data and parameters interface from TRD subsystem used in L1 tracker (declaration)
 * @since  31.05.2022
 * @author S.Zharko <s.zharko@gsi.de>
 ***************************************************************************************************/

#ifndef CbmTrdTrackingInterface_h
#define CbmTrdTrackingInterface_h 1

#include "CbmHit.h"
#include "CbmTrackingDetectorInterfaceBase.h"
#include "CbmTrdAddress.h"
#include "CbmTrdParModDigi.h"
#include "CbmTrdParSetDigi.h"

#include "FairTask.h"

#include "TMath.h"

#include <iostream>
#include <vector>

/// @class CbmTrdTrackingInterface
/// @brief A CbmL1 subtask, which provides necessary methods for CA tracker to access the geometry and dataflow s
/// ettings.
///
/// @note For TRD one tracking station is a TRD module!
///
class CbmTrdTrackingInterface : public FairTask, public CbmTrackingDetectorInterfaceBase {
public:
  /// @brief Default constructor
  CbmTrdTrackingInterface();

  /// @brief Destructor
  ~CbmTrdTrackingInterface();

  /// @brief Copy constructor
  CbmTrdTrackingInterface(const CbmTrdTrackingInterface&) = delete;

  /// @brief Move constructor
  CbmTrdTrackingInterface(CbmTrdTrackingInterface&&) = delete;

  /// @brief Copy assignment operator
  CbmTrdTrackingInterface& operator=(const CbmTrdTrackingInterface&) = delete;

  /// @brief Move assignment operator
  CbmTrdTrackingInterface& operator=(CbmTrdTrackingInterface&&) = delete;

  /// @brief Gets name of this subsystem
  std::string GetDetectorName() const { return "TRD"; }

  /// @brief Gets actual number of tracking stations, provided by the current geometry setup
  int GetNtrackingStations() const;

  /// @brief  Gets stereo angles of the two independent measured coordinates
  /// @note   The tracking does not use this method. It is only used by the QA task.
  /// @param  address  detector unique identifier
  /// @return [phiU, phiV] - Stereo angles [rad]
  std::tuple<double, double> GetStereoAnglesSensor(int address) const;

  /// @brief  Gets a tracking station of a CbmHit
  /// @param  hit  A pointer to CbmHit
  /// @return Local index of the tracking station
  int GetTrackingStationIndex(const CbmHit* hit) const { return GetTrackingStationIndex(hit->GetAddress()); }

  /// @brief  Gets a tracking station by the address
  /// @param  address  Unique element address
  /// @return Local index of the tracking station
  int GetTrackingStationIndex(int address) const { return CbmTrdAddress::GetLayerId(address); }

  /// @brief  Gets a tracking station of a FairMCPoint
  /// @param  point  A pointer to FairMCHit
  /// @return Local index of the tracking station
  int GetTrackingStationIndex(const FairMCPoint* point) const
  {
    return GetTrackingStationIndex(point->GetDetectorID());
  }

  /// @brief  Gets upper edge of a station along the X-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the X-axis [cm]
  double GetXmax(int stationId) const { return GetTrdModulePar(stationId)->GetSizeX(); }

  /// @brief  Gets lower edge coordinate of a station along the X-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the X-axis [cm]
  double GetXmin(int stationId) const { return -GetTrdModulePar(stationId)->GetSizeX(); }

  /// @brief  Gets upper edge of a station along the Y-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the Y-axis [cm]
  double GetYmax(int stationId) const { return GetTrdModulePar(stationId)->GetSizeY(); }

  /// @brief  Gets lower edge of a station along the Y-axis
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Size of station along the Y-axis [cm]
  double GetYmin(int stationId) const { return -GetTrdModulePar(stationId)->GetSizeY(); }

  /// @brief  Gets reference z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Reference z position of the station [cm]
  double GetZref(int stationId) const { return GetTrdModulePar(stationId)->GetZ(); }

  /// @brief  Gets max z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return max Z of the station [cm]
  double GetZmax(int stationId) const { return GetZref(stationId) + GetTrdModulePar(stationId)->GetSizeZ(); }

  /// @brief  Gets min z of the station
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return min Z of the station [cm]
  double GetZmin(int stationId) const { return GetZref(stationId) - GetTrdModulePar(stationId)->GetSizeZ(); }

  /// @brief  Check if station provides time measurements
  /// @param  stationId  Tracking station ID in the setup (NOTE: must be in range [0..GetNstations()-1])
  /// @return Flag: true - station provides time measurements, false - station does not provide time measurements
  bool IsTimeInfoProvided(int /*stationId*/) const { return true; }

  /// @brief  Gets x,y,t ranges of a CbmTrdHit
  /// @param  hit  A hit
  /// @return range X, Y, T
  std::tuple<double, double, double> GetHitRanges(const CbmPixelHit& hit) const;

  /// @brief FairTask: Init method
  InitStatus Init();

  /// @brief Gets pointer to the instance of the CbmTrdTrackingInterface
  static CbmTrdTrackingInterface* Instance() { return fpInstance; }

  /// @brief FairTask: ReInit method
  InitStatus ReInit();

  /// @brief  FairTask: sets parameter containers up
  void SetParContainers();

private:
  /// @brief  Gets pointer to the TRD module
  /// @param  moduleId  Id of the Trd module
  /// @return Pointer to the particular CbmTrdParModDigi object
  __attribute__((always_inline)) CbmTrdParModDigi* GetTrdModulePar(int moduleId) const
  {
    return static_cast<CbmTrdParModDigi*>(fTrdDigiPar->GetModulePar(fTrdDigiPar->GetModuleId(moduleId)));
  }

  static CbmTrdTrackingInterface* fpInstance;  ///< Instance of the class

  CbmTrdParSetDigi* fTrdDigiPar {nullptr};
  //CbmTrdParModDigi* fTrdModuleInfo {nullptr};

  ClassDef(CbmTrdTrackingInterface, 0);
};

#endif  // CbmTrdTrackingInterface
