/* Copyright (C) 2021-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/************************************************************************************************************
 * @file L1BaseStationInfo.h
 * @brief A base class for a L1 station interface
 * @since 18.12.2021
 *
 * The class is implemented to connect concrete experimental setup (CBM or BMN in CbmL1 
 * or BmnL1) with general L1Algo tracking.
 * 
 ***********************************************************************************************************/

#ifndef L1BaseStationInfo_h
#define L1BaseStationInfo_h 1

// L1 Core
#include "CaMaterialMap.h"
#include "CaObjectInitController.h"
#include "CaSimd.h"
#include "CaStation.h"

// C++ std
#include <bitset>
#include <functional>
#include <string>

using namespace cbm::algo;      // TODO: remove this line
using namespace cbm::algo::ca;  // TODO: remove this line

enum class L1DetectorID;

/// A base class which provides interface to L1Algo station geometry
class L1BaseStationInfo {
  /// Enumeration for internal logic control
  enum class EManagementFlag
  {
    kThicknessMapMoved,  ///< if the thickness map was moved from the L1BaseStationInfo instance
    kEnd
  };
  using ManagementFlags_t = std::bitset<static_cast<int>(EManagementFlag::kEnd)>;

public:
  /// Enumeration of fields, which must be initialized so the object can pass the threshold
  enum class EInitKey
  {
    // Basic fields initialization
    kDetectorID,      ///< detector ID
    kStationID,       ///< station ID
    kTrackingStatus,  ///< flag, if station is used in tracking or not
    kXmax,            ///< max size in X direction
    kYmax,            ///< max size in Y direction
    // ca::Station initialization
    kType,          ///< station type
    kTimeInfo,      ///< if time info is used (flag)
    kFieldStatus,   ///< if station is placed in field (flag)
    kZref,          ///< reference z of the station
    kZmin,          ///< min z of the station
    kZmax,          ///< max z of the station
    kThicknessMap,  ///< thickness map of the station (optional?)
    kFieldSlice,    ///< ca::Station.ca::FieldSlice object initialization
    // The last item is equal to the number of bits in fInitFlags
    kEnd
  };

  using InitController_t = ca::ObjectInitController<static_cast<int>(EInitKey::kEnd), EInitKey>;

  //
  // CONSTRUCTORS AND DESTRUCTORS
  //
  /// Default constructor
  L1BaseStationInfo() = default;

  /// Constructor from stationID and detectorID
  L1BaseStationInfo(L1DetectorID detetorID, int stationID) noexcept;

  /// Destructor
  ~L1BaseStationInfo() noexcept = default;

  /// Copy constructor
  L1BaseStationInfo(const L1BaseStationInfo& other) = default;

  /// Less operator for L1BaseStationInfo object to sort stations in Z
  bool operator<(const L1BaseStationInfo& right) const { return (GetZref() < right.GetZref()); }

  /// Gets detector ID
  L1DetectorID GetDetectorID() const { return fDetectorID; }

  /// Gets a coefficient with idx for the field x-component approximation
  fvec GetFieldSliceCx(int idx) const { return fL1Station.fieldSlice.cx[idx]; }

  /// Gets a coefficient with idx for the field y-component approximation
  fvec GetFieldSliceCy(int idx) const { return fL1Station.fieldSlice.cy[idx]; }

  /// Gets a coefficient with idx for the field z-component approximation
  fvec GetFieldSliceCz(int idx) const { return fL1Station.fieldSlice.cz[idx]; }

  /// Gets array of the coefficients for the field x-component approximation
  const fvec* GetFieldSliceCx() const { return fL1Station.fieldSlice.cx; }

  /// Gets array of the coefficients for the field y-component approximation
  const fvec* GetFieldSliceCy() const { return fL1Station.fieldSlice.cy; }

  /// Gets array of the coefficients for the field z-component approximation
  const fvec* GetFieldSliceCz() const { return fL1Station.fieldSlice.cz; }


  /// Gets field status: 0 - station is outside the field, 1 - station is inside the field
  int GetFieldStatus() const { return fL1Station.fieldStatus; }

  /// Gets a const reference to the ca::ObjectInitController object
  const InitController_t& GetInitController() const { return fInitController; }

  /// Gets a reference to ca::Station info field of the L1BaseStation info
  const ca::Station& GetL1Station() const;

  /// Gets a reference to ca::MaterialMap map
  const ca::MaterialMap& GetMaterialMap() const;

  /// Gets station ID
  int GetStationID() const { return fStationID; }

  /// Gets station type
  int GetStationType() const { return fL1Station.type; }

  /// Gets tracking status: true - station is active for tracking, false - station exists, but not used in tracking
  bool GetTrackingStatus() const { return fTrackingStatus; }

  /// Gets maximum distance between station center and its edge in x direction
  double GetXmax() const { return fXmax; }

  /// Gets maximum distance between station center and its edge in y direction
  double GetYmax() const { return fYmax; }

  /// Gets double precised z position of the station [cm]
  double GetZref() const { return fZref; }

  /// Gets min z of the station [cm]
  double GetZmin() const { return fZmin; }

  /// Gets max z of the station [cm]
  double GetZmax() const { return fZmax; }

  /// Gets SIMD vectorized z position of the station [cm]
  fvec GetZsimdVec() const { return fL1Station.fZ; }

  /// Prints registered fields
  /// verbosity = 0: print only station id, detector id and address in one line
  /// verbosity = 1: print basic ca::Station fields
  /// verbosity = 2: print all L1Staion fields
  void Print(int verbosity = 0) const;

  /// Resets fields to the default values
  void Reset();

  /// Sets detector ID
  void SetDetectorID(L1DetectorID inID);

  /// Sets flag: true - station is placed in field, false - station is placed outside the field
  void SetFieldStatus(int fieldStatus);

  /// Sets arrays of the approcimation
  /// \param getField A user function, which gets a xyz array of position coordinates and fills B array
  ///                 of magnetic field components in position
  void SetFieldFunction(const std::function<void(const double (&xyz)[3], double (&B)[3])>& getFieldValue);

  /// Sets station thickness in units of radiation length mapped vs. position in XY plane (copy semantics)
  /// \param thicknessMap  Map of station thickness in units of radiation length
  void SetMaterialMap(const ca::MaterialMap& thicknessMap);

  /// Sets station thickness in units of radiation length mapped vs. position in XY plane (move semantics)
  /// \param thicknessMap  Map of station thickness in units of radiation length
  void SetMaterialMap(ca::MaterialMap&& thicknessMap) noexcept;

  /// Sets station ID
  [[deprecated("Please, use constructor to set station ID")]] void SetStationID(int inID);

  /// Sets type of station
  void SetStationType(int inType);  // TODO: this is a temporary solution (S.Zh.)

  /// Sets flag: 0 - time information is not provided by this detector type
  ///            1 - time information is provided by the detector and can be used in tracking
  void SetTimeInfo(int inTimeInfo);

  /// Sets tracking status: true - station is active for tracking, false - station exists, but not used in tracking
  void SetTrackingStatus(bool flag);

  /// Sets maximum distance between station center and its edge in x direction
  void SetXmax(double aSize);

  /// Sets maximum distance between station center and its edge in y direction
  void SetYmax(double aSize);

  /// Sets reference z position of the station
  void SetZref(double inZ);

  /// Sets min z of the station
  void SetZmin(double inZ);

  /// Sets max z of the station
  void SetZmax(double inZ);

  /// Swap method for easy implementation of move constructor and copy and move assignment operator
  void Swap(L1BaseStationInfo& other) noexcept;

  /// Takes an instance of ca::MaterialMap map
  /// This method can be called only once. It will catch the content of ca::MaterialMap map (using move semantics), and after the warning will be generated
  /// both for GetMaterialMap and TakeMaterialMap
  ca::MaterialMap&& TakeMaterialMap();

  /// String representation of class contents
  /// \param indentLevel  Number of indent characters in the output
  std::string ToString(int verbosityLevel = 0, int indentLevel = 0) const;

private:
  L1DetectorID fDetectorID {static_cast<L1DetectorID>(0)};  ///< Detector ID
  int fStationID {-1};                                      ///< Local ID of a station
  bool fTrackingStatus {false};                             ///< Tracking status: true - station is used for tracking
  double fXmax {0};                       ///< Maximum distance between station center and its edge in x direction
  double fYmax {0};                       ///< Maximum distance between station center and its edge in y direction
  double fZref {0};                       ///< reference z
  double fZmin {0};                       ///< min z
  double fZmax {0};                       ///< max z
  ca::Station fL1Station {};              ///< ca::Station structure, describes a station in L1Algo
  ca::MaterialMap fThicknessMap {};       ///< Map of station thickness in units of radiation length
  InitController_t fInitController {};    ///< Class fileds initialization flags
  ManagementFlags_t fManagementFlags {};  ///< bitset flags to manage internal behaviour of the class
};

/// swap function for two L1BaseStationInfo objects, expected to be used instead of std::swap
inline void swap(L1BaseStationInfo& a, L1BaseStationInfo& b) noexcept { a.Swap(b); }

#endif  // L1BaseStationInfo_h
