/* Copyright (C) 2009-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Ivan Kisel, Sergey Gorbunov [committer], Igor Kulakov, Maksym Zyzak */

/*
 *=====================================================
 *
 *  CBM Level 1 4D Reconstruction
 *
 *  Authors: V.Akishina, I.Kisel,  S.Gorbunov, I.Kulakov, M.Zyzak
 *  Documentation: V.Akishina
 *
 *  e-mail : v.akishina@gsi.de
 *
 *=====================================================
 *
 *  Finds tracks using the Cellular Automaton algorithm
 *
 */


#include "CbmL1.h"
#include "CbmL1MCTrack.h"

#include "CaTrack.h"
#include "CaTrackParam.h"
#include "L1Algo.h"
#include "L1Assert.h"
#include "L1Branch.h"
#include "L1Fit.h"
#include "L1Grid.h"
#include "L1HitArea.h"
#include "L1HitPoint.h"
#include "L1TripletConstructor.h"
#ifdef DRAW
#include "utils/L1AlgoDraw.h"
#endif  // DRAW
#ifdef PULLS
#include "utils/L1AlgoPulls.h"
#endif  // PULLS
#ifdef TRIP_PERFORMANCE
#include "utils/L1AlgoEfficiencyPerformance.h"
#endif  // TRIP_PERFORMANCE
#ifdef DOUB_PERFORMANCE
#include "utils/L1AlgoEfficiencyPerformance.h"
#endif  // DOUB_PERFORMANCE

#include <algorithm>
#include <fstream>
#include <iostream>
#include <list>
#include <map>

#include <stdio.h>

#include "CaToolsDebugger.h"


// using namespace std;
using std::cout;
using std::endl;
using Track = cbm::algo::ca::Track;

bool L1Algo::checkTripletMatch(const L1Triplet& l, const L1Triplet& r, fscal& dchi2) const
{
  dchi2 = 1.e20;

  if (r.GetMHit() != l.GetRHit()) return false;
  if (r.GetLHit() != l.GetMHit()) return false;

  if (r.GetMSta() != l.GetRSta()) return false;
  if (r.GetLSta() != l.GetMSta()) return false;


  if (r.IsMomentumFitted()) {
    assert(l.IsMomentumFitted());

    fscal dqp = l.GetQp() - r.GetQp();
    fscal Cqp = l.GetCqp() + r.GetCqp();

    if (!std::isfinite(dqp)) return false;
    if (!std::isfinite(Cqp)) return false;

    if (dqp * dqp > fTripletLinkChi2 * Cqp) {
      return false;  // bad neighbour // CHECKME why do we need recheck it?? (it really change result)
    }
    dchi2 = dqp * dqp / Cqp;
  }
  else {
    fscal dtx = l.GetTx() - r.GetTx();
    fscal Ctx = l.GetCtx() + r.GetCtx();

    fscal dty = l.GetTy() - r.GetTy();
    fscal Cty = l.GetCty() + r.GetCty();

    // it shouldn't happen, but happens sometimes

    if (!std::isfinite(dtx)) return false;
    if (!std::isfinite(dty)) return false;
    if (!std::isfinite(Ctx)) return false;
    if (!std::isfinite(Cty)) return false;

    if (dty * dty > fTripletLinkChi2 * Cty) return false;
    if (dtx * dtx > fTripletLinkChi2 * Ctx) return false;

    //dchi2 = 0.5f * (dtx * dtx / Ctx + dty * dty / Cty);
    dchi2 = 0.;
  }

  if (!std::isfinite(dchi2)) return false;

  return true;
}


// **************************************************************************************************
// *                                                                                                *
// *                            ------ CATrackFinder procedure ------                               *
// *                                                                                                *
// **************************************************************************************************

void L1Algo::CaTrackFinderSlice()
{
  fNFindIterations = fParameters.GetNcaIterations();

  /*******************************/ /**
   * Performance monitors setting
   **********************************/

#ifdef PULLS
  static L1AlgoPulls* l1Pulls_ = new L1AlgoPulls();
  fL1Pulls                     = l1Pulls_;
  fL1Pulls->Init();
#endif
#ifdef TRIP_PERFORMANCE
  static L1AlgoEfficiencyPerformance<3>* l1Eff_triplets_ = new L1AlgoEfficiencyPerformance<3>();
  fL1Eff_triplets                                        = l1Eff_triplets_;
  fL1Eff_triplets->Init();
  static L1AlgoEfficiencyPerformance<3>* l1Eff_triplets2_ = new L1AlgoEfficiencyPerformance<3>();
  fL1Eff_triplets2                                        = l1Eff_triplets2_;
  fL1Eff_triplets2->Init();
#endif
#ifdef DOUB_PERFORMANCE
  static L1AlgoEfficiencyPerformance<2>* l1Eff_doublets_ = new L1AlgoEfficiencyPerformance<2>();
  fL1Eff_doublets                                        = l1Eff_doublets_;
  fL1Eff_doublets->Init();
#endif

#ifdef DRAW
  if (!draw) draw = new L1AlgoDraw;
  draw->InitL1Draw(this);
#endif


#if defined(XXX) || defined(COUNTERS)
  static unsigned int stat_N = 0;  // number of events
  stat_N++;
#endif

#ifdef COUNTERS
  static Tindex stat_nStartHits              = 0;
  static Tindex stat_nHits[fNFindIterations] = {0};

  static Tindex stat_nSinglets[fNFindIterations] = {0};
  //  static Tindex stat_nDoublets[fNFindIterations] = {0};
  static Tindex stat_nTriplets[fNFindIterations] = {0};

  static Tindex stat_nLevels[constants::size::MaxNstations - 2][fNFindIterations] = {{0}};
  static Tindex stat_nCalls[fNFindIterations]                                     = {0};  // n calls of CAFindTrack
  static Tindex stat_nTrCandidates[fNFindIterations]                              = {0};

  for (int iS = 0; iS < fParameters.GetNstationsActive(); ++iS) {
    stat_nStartHits += fSliceHitIds[iS].size();
  }
#endif

  /********************************/ /**
   * CATrackFinder routine setting
   ***********************************/

  ResetSliceData();


  ca::HitIndex_t nGridHitsFilled = 0;
  for (int iS = 0; iS < fParameters.GetNstationsActive(); ++iS) {

    fMaxRangeX[iS]         = 0.;
    fMaxRangeY[iS]         = 0.;
    fMaxRangeT[iS]         = 0.;
    bool timeUninitialised = 1;
    fscal lasttime         = 0;
    fscal starttime        = 0;


    fscal gridMinX = -0.1;
    fscal gridMaxX = 0.1;
    fscal gridMinY = -0.1;
    fscal gridMaxY = 0.1;

    for (ca::HitIndex_t ih = 0; ih < fSliceHitIds[iS].size(); ++ih) {
      const ca::Hit& h = fInputData.GetHit(fSliceHitIds[iS][ih]);

      if (fMaxRangeX[iS] < h.rangeX) { fMaxRangeX[iS] = h.rangeX; }
      if (fMaxRangeY[iS] < h.rangeY) { fMaxRangeY[iS] = h.rangeY; }
      if (fMaxRangeT[iS] < h.rangeT) { fMaxRangeT[iS] = h.rangeT; }

      auto [x, y] = GetHitCoorOnGrid(h);
      if (x < gridMinX) { gridMinX = x; }
      if (x > gridMaxX) { gridMaxX = x; }
      if (y < gridMinY) { gridMinY = y; }
      if (y > gridMaxY) { gridMaxY = y; }

      const fscal time = h.t;
      assert(std::isfinite(time));
      if (timeUninitialised || lasttime < time) { lasttime = time; }
      if (timeUninitialised || starttime > time) { starttime = time; }
      timeUninitialised = false;
    }

    // TODO: changing the grid also changes the result. Investigate why it happens.
    // TODO: find the optimal grid size

    /*  SG: old code from Valentina
      int nSliceHits = fSliceHitIds.size();
      int nSliceHits = fSliceHitIdsStopIndex[iS] - fSliceHitIdsStartIndex[iS];
      fscal yStep    = 1.5 / sqrt(nSliceHits + 1);  // empirics. 0.01*sqrt(2374) ~= 0.5
      if (yStep > 0.3) yStep = 0.3;
      if (yStep < 0.01) yStep = 0.01;
      fscal xStep = yStep * 3;
      vGridTime[iS].BuildBins(-1, 1, -0.6, 0.6, starttime, lasttime, xStep, yStep, (lasttime - starttime + 1));
    */

    int nSliceHits = fSliceHitIds[iS].size();
    fscal sizeY    = gridMaxY - gridMinY;
    fscal sizeX    = gridMaxX - gridMinX;
    int nBins2D    = 1 + nSliceHits;
    fscal yStep    = sizeY / sqrt(nBins2D);
    fscal xStep    = sizeX / sqrt(nBins2D);
    if (yStep > 0.3) yStep = 0.3;
    if (yStep < 0.01) yStep = 0.01;
    if (xStep > 0.3) xStep = 0.3;
    if (xStep < 0.01) xStep = 0.01;

    vGridTime[iS].BuildBins(gridMinX, gridMaxX, gridMinY, gridMaxY, starttime, lasttime, xStep, yStep,
                            (lasttime - starttime + 1));
    vGridTime[iS].StoreHits(*this, iS, nGridHitsFilled);
  }

  fGridHits.reduce(nGridHitsFilled);
  fGridHitsBuf.reduce(nGridHitsFilled);
  fGridHitIds.reduce(nGridHitsFilled);
  fGridHitIdsBuf.reduce(nGridHitsFilled);
  fGridPoints.reduce(nGridHitsFilled);
  fGridPointsBuf.reduce(nGridHitsFilled);


  for (unsigned int ih = 0; ih < fGridHits.size(); ++ih) {
    fGridPoints[ih].Set(fGridHits[ih]);
  }

  /********************************/ /**
   *  Loop over tracking iterations
   ***********************************/


  // ---- Loop over Track Finder iterations ----------------------------------------------------------------//


  L1ASSERT(0, fNFindIterations == (int) fParameters.GetCAIterations().size());
  isec = 0;  // TODO: temporary! (S.Zharko)

  for (const auto& caIteration : fParameters.GetCAIterations())  // all finder
  {
    fpCurrentIteration = &caIteration;  // select current iteration

    for (int j = 0; j < fParameters.GetNstationsActive(); j++) {
      fTriplets[j].clear();
    }
    /// isec - number of current iterations, fNFindIterations - number of all iterations
#ifdef COUNTERS
    Tindex nSinglets = 0;
#endif

    if (isec != 0) {

      fGridHitIds.swap(fGridHitIdsBuf);
      fGridHits.swap(fGridHitsBuf);
      fGridPoints.swap(fGridPointsBuf);
    }
    // TODO: Replace NStations with fInitManager.GetNstationsGeom() (S.Zharko)
    for (int ist = 0; ist < fParameters.GetNstationsActive(); ++ist) {
      for (ca::HitIndex_t ih = fGridHitStartIndex[ist]; ih < fGridHitStopIndex[ist]; ++ih) {
        //SG!!
        fHitFirstTriplet[ih] = 0;
        fHitNtriplets[ih]    = 0;
      }
    }
    /*
   fHitFirstTriplet.assign(fGridHitStopIndex[fParameters.GetNstationsActive() - 1],0);
   fHitNtriplets.assign(fGridHitStopIndex[fParameters.GetNstationsActive() - 1],0);
*/
    {
      // #pragma omp  task
      {
        // --- SET PARAMETERS FOR THE ITERATION ---

        fFirstCAstation      = caIteration.GetFirstStationIndex();
        fTrackChi2Cut        = caIteration.GetTrackChi2Cut();
        fDoubletChi2Cut      = caIteration.GetDoubletChi2Cut();  //11.3449 * 2.f / 3.f;  // prob = 0.1
        fTripletChi2Cut      = caIteration.GetTripletChi2Cut();  //21.1075;  // prob = 0.01%
        fTripletFinalChi2Cut = caIteration.GetTripletFinalChi2Cut();
        fPickGather          = caIteration.GetPickGather();       //3.0;
        fTripletLinkChi2     = caIteration.GetTripletLinkChi2();  //5.0;
        fMaxInvMom           = caIteration.GetMaxQp();            //1.0 / 0.5;  // max considered q/p
        fMaxSlopePV          = caIteration.GetMaxSlopePV();       //1.1;
        fMaxSlope            = caIteration.GetMaxSlope();         //2.748;  // corresponds to 70 grad

        // define the target
        fTargX = fParameters.GetTargetPositionX();
        fTargY = fParameters.GetTargetPositionY();
        fTargZ = fParameters.GetTargetPositionZ();

        fscal SigmaTargetX = caIteration.GetTargetPosSigmaX();
        fscal SigmaTargetY = caIteration.GetTargetPosSigmaY();  // target constraint [cm]

        // Select magnetic field. For primary tracks - fVtxFieldValue, for secondary tracks - st.fieldSlice
        if (caIteration.GetPrimaryFlag()) { fTargB = fParameters.GetVertexFieldValue(); }
        else {
          fTargB = fParameters.GetStation(0).fieldSlice.GetFieldValue(0, 0);
        }  // NOTE: calculates field fTargB in the center of 0th station

        fIsTargetField = (fabs(fTargB.y[0]) > 0.001);

        TargetXYInfo.C00 = SigmaTargetX * SigmaTargetX;
        TargetXYInfo.C10 = 0;
        TargetXYInfo.C11 = SigmaTargetY * SigmaTargetY;

        /// Set correction in order to take into account overlaping and iff z.
        /// The reason is that low momentum tracks are too curved and goes not from target direction. That's why sort by hit_y/hit_z is not work idealy
        /// If sort by y then it is max diff between same station's modules (~0.4cm)
        fMaxDZ = caIteration.GetMaxDZ();  //0;
      }

#ifndef L1_NO_ASSERT
      for (int istal = fParameters.GetNstationsActive() - 1; istal >= 0; istal--) {
        L1_ASSERT(fGridHitStopIndex[istal] >= fGridHitStartIndex[istal],
                  fGridHitStopIndex[istal] << " >= " << fGridHitStartIndex[istal]);
        L1_ASSERT(fGridHitStopIndex[istal] <= fGridHits.size(), fGridHitStopIndex[istal] << " <= " << fGridHits.size());
      }
#endif  // L1_NO_ASSERT
    }


    ///   stage for triplets creation

    L1TripletConstructor constructor1(this);
    L1TripletConstructor constructor2(this);
    L1TripletConstructor constructor3(this);

    for (int istal = fParameters.GetNstationsActive() - 2; istal >= fFirstCAstation;
         istal--) {  // start with downstream chambers
      Tindex NHits_l = fGridHitStopIndex[istal] - fGridHitStartIndex[istal];
      for (Tindex ih = 0; ih < NHits_l; ++ih) {
        constructor1.CreateTripletsForHit(istal, istal + 1, istal + 2, ih);
        if (fpCurrentIteration->GetJumpedFlag()) {
          constructor2.CreateTripletsForHit(istal, istal + 1, istal + 3, ih);
          constructor3.CreateTripletsForHit(istal, istal + 2, istal + 3, ih);
        }

        auto oldSize = fTriplets[istal].size();

        fTriplets[istal].insert(fTriplets[istal].end(), constructor1.GetTriplets().begin(),
                                constructor1.GetTriplets().end());
        fTriplets[istal].insert(fTriplets[istal].end(), constructor2.GetTriplets().begin(),
                                constructor2.GetTriplets().end());
        fTriplets[istal].insert(fTriplets[istal].end(), constructor3.GetTriplets().begin(),
                                constructor3.GetTriplets().end());

        const ca::HitIndex_t ihitl = ih + fGridHitStartIndex[istal];
        fHitFirstTriplet[ihitl]  = PackTripletId(istal, oldSize);
        fHitNtriplets[ihitl]     = fTriplets[istal].size() - oldSize;
      }
    }  // istal


    // search for neighbouring triplets

    for (int istal = fParameters.GetNstationsActive() - 2; istal >= fFirstCAstation;
         istal--) {  // start with downstream chambers

      for (unsigned int it = 0; it < fTriplets[istal].size(); ++it) {

        L1Triplet& tr = fTriplets[istal][it];
        tr.SetLevel(0);
        tr.SetFNeighbour(0);
        tr.SetNNeighbours(0);

        if (istal > (fParameters.GetNstationsActive() - 4)) break;

        unsigned int nNeighbours = fHitNtriplets[tr.GetMHit()];

        unsigned int neighLocation = fHitFirstTriplet[tr.GetMHit()];
        unsigned int neighStation  = TripletId2Station(neighLocation);
        unsigned int neighTriplet  = TripletId2Triplet(neighLocation);

        if (nNeighbours > 0) { assert((int) neighStation == istal + 1 || (int) neighStation == istal + 2); }

        unsigned char level = 0;

        for (unsigned int iN = 0; iN < nNeighbours; ++iN, ++neighTriplet, ++neighLocation) {

          L1Triplet& neighbour = fTriplets[neighStation][neighTriplet];

          fscal dchi2 = 0.;
          if (!checkTripletMatch(tr, neighbour, dchi2)) continue;

          if (tr.GetFNeighbour() == 0) { tr.SetFNeighbour(neighLocation); }

          tr.SetNNeighbours(neighLocation - tr.GetFNeighbour() + 1);

          const unsigned char neighbourLevel = neighbour.GetLevel();

          if (level <= neighbourLevel) { level = neighbourLevel + 1; }
        }
        tr.SetLevel(level);
      }  // neighbour search
    }    // istal

    ///====================================================================
    ///=                                                                  =
    ///=        Collect track candidates. CREATE TRACKS                   =
    ///=                                                                  =
    ///====================================================================

    // min level to start triplet. So min track length = min_level+3.
    int min_level = std::min(caIteration.GetMinNhits(), caIteration.GetMinNhitsStation0()) - 3;

    // TODO: SZh 04.10.2022: Why this fatal error is called?
    // NOTE: This line was wrapped into TRACKS_FROM_TRIPLETS ifdef
    //LOG(FATAL) << "L1CATrackFinder: min_level undefined in " << __FILE__ << " : " << __LINE__;
    if (fpCurrentIteration->GetTrackFromTripletsFlag()) { min_level = 0; }

    L1Branch curr_tr;
    L1Branch new_tr[constants::size::MaxNstations];
    L1Branch best_tr;
    fscal curr_chi2 = 0;

    fscal best_chi2      = 0;
    unsigned char best_L = 0;

    unsigned char curr_L = 1;
    int ndf              = 1;


    // collect consequtive: the longest tracks, shorter, more shorter and so on
    for (int firstTripletLevel = fParameters.GetNstationsActive() - 3; firstTripletLevel >= min_level;
         firstTripletLevel--) {
      // choose length in triplets number - firstTripletLevel - the maximum possible triplet level among all triplets in the searched candidate

      //  how many levels to check
      int nlevel = (fParameters.GetNstationsActive() - 2) - firstTripletLevel + 1;

      const unsigned char min_best_l =
        (firstTripletLevel > min_level) ? firstTripletLevel + 2 : min_level + 3;  // loose maximum


      fTrackCandidates.clear();

      fStripToTrack.reset(fInputData.GetNhitKeys(), -1);

      //== Loop over triplets with the required level, find and store track candidates

      for (int istaF = fFirstCAstation; istaF <= fParameters.GetNstationsActive() - 3 - firstTripletLevel; ++istaF) {

        if (--nlevel == 0) break;  //TODO: SG: this is not needed

        for (Tindex itrip = 0; itrip < (Tindex) fTriplets[istaF].size(); ++itrip) {

          L1Triplet& first_trip = (fTriplets[istaF][itrip]);
          if (fvHitKeyFlags[fGridHits[first_trip.GetLHit()].f] || fvHitKeyFlags[fGridHits[first_trip.GetLHit()].b]) {
            continue;
          }
          //               ghost suppression !!!
          //

          if (!fpCurrentIteration->GetTrackFromTripletsFlag()) {  // ghost suppression !!!
            int nHits = 3 + first_trip.GetLevel();
            if (fGridHits[first_trip.GetLHit()].iSt == 0) {
              if (nHits < fpCurrentIteration->GetMinNhitsStation0()) { continue; }
            }
            else {
              if (nHits < fpCurrentIteration->GetMinNhits()) { continue; }
            }
          }

          // Collect triplets, which can start a track with length equal to firstTipletLevel + 3. This cut suppresses
          // ghost tracks, but does not affect the efficiency
          if (first_trip.GetLevel() < firstTripletLevel) { continue; }

          //  curr_tr.Momentum = 0.f;
          curr_tr.chi2 = 0.f;
          //   curr_tr.Lengtha  = 0;
          curr_tr.ista = 0;
          curr_tr.fHits.clear();
          curr_tr.fHits.push_back(fGridHitIds[first_trip.GetLHit()]);
          curr_tr.NHits = 1;

          curr_L    = 1;
          curr_chi2 = first_trip.GetChi2();

          best_tr   = (curr_tr);
          best_chi2 = curr_chi2;
          best_L    = curr_L;

          CAFindTrack(istaF, best_tr, best_L, best_chi2, &first_trip, (curr_tr), curr_L, curr_chi2, min_best_l,
                      new_tr);  /// reqursive func to build a tree of possible track-candidates and choose the best

          if (best_L < firstTripletLevel + 2) continue;  // loose maximum one hit

          if (best_L < min_level + 3) continue;  // should find all hits for min_level

          ndf = best_L * 2 - 5;

          // TODO: automatize the NDF calculation

          if (kGlobal == fTrackingMode || kMcbm == fTrackingMode) { ndf = best_L * 2 - 4; }

          best_chi2 = best_chi2 / ndf;  //normalize

          if (fGhostSuppression) {
            if (3 == best_L) {
              if (!fpCurrentIteration->GetPrimaryFlag() && (istaF != 0)) continue;  // too /*short*/ non-MAPS track
              if (fpCurrentIteration->GetPrimaryFlag() && (best_chi2 > 5.0)) continue;
            }
          }
          fTrackCandidates.push_back(best_tr);
          L1Branch& tr = fTrackCandidates.back();
          tr.Set(istaF, best_L, best_chi2, first_trip.GetQp(), fTrackCandidates.size() - 1);
          if (0) {  // debug
            cout << "track " << fTrackCandidates.size() - 1 << " found, L = " << (int) best_L << " chi2= " << best_chi2
                 << endl;
            cout << " hits: ";
            for (unsigned int i = 0; i < tr.fHits.size(); i++) {
              cout << GetMcTrackIdForCaHit(tr.fHits[i]) << " ";
            }
            cout << endl;
          }
        }  // itrip
      }    // istaF

      for (Tindex iTrack = 0; iTrack < (Tindex) fTrackCandidates.size(); ++iTrack) {
        L1Branch& tr = fTrackCandidates[iTrack];
        tr.fIsAlive  = false;
      }

      bool repeatCompetition = true;
      int iComp              = 0;
      for (iComp = 0; (iComp < 100) && repeatCompetition; ++iComp) {

        // == Loop over track candidates and mark their strips

        repeatCompetition = false;

        for (Tindex iTrack = 0; iTrack < (Tindex) fTrackCandidates.size(); ++iTrack) {
          L1Branch& tr = fTrackCandidates[iTrack];
          if (tr.fIsAlive) continue;

          for (int iHit = 0; iHit < (int) tr.fHits.size(); ++iHit) {
            const ca::Hit& h = fInputData.GetHit(tr.fHits[iHit]);
            bool isAlive   = true;
            {  // front  strip
              auto& stripF = (fStripToTrack)[h.f];
              if ((stripF >= 0) && (stripF != tr.fID)) {  // strip is used by other candidate
                const auto& other = fTrackCandidates[stripF];
                if (!other.fIsAlive && L1Branch::compareCand(tr, other)) { stripF = tr.fID; }
                else {
                  isAlive = false;
                }
              }
              else {
                stripF = tr.fID;
              }
              if (!isAlive) { break; }
            }

            {  // back strip
              auto& stripB = (fStripToTrack)[h.b];
              if ((stripB >= 0) && (stripB != tr.fID)) {  // strip is used by other candidate
                const auto& other = fTrackCandidates[stripB];
                if (!other.fIsAlive && L1Branch::compareCand(tr, other)) { stripB = tr.fID; }
                else {
                  isAlive = false;
                }
              }
              else {
                stripB = tr.fID;
              }
              if (!isAlive) { break; }
            }
          }  // loop over hits
        }    // itrack

        // == Loop over track candidates and select those that are alive

        for (Tindex iTrack = 0; iTrack < (Tindex) fTrackCandidates.size(); ++iTrack) {
          L1Branch& tr = fTrackCandidates[iTrack];
          if (tr.fIsAlive) { continue; }

          tr.fIsAlive = true;
          for (int iHit = 0; tr.fIsAlive && (iHit < (int) tr.fHits.size()); ++iHit) {
            const ca::Hit& h = fInputData.GetHit(tr.fHits[iHit]);
            tr.fIsAlive    = tr.fIsAlive && ((fStripToTrack)[h.f] == tr.fID) && ((fStripToTrack)[h.b] == tr.fID);
          }

          if (!tr.fIsAlive) {  // release strips
            for (int iHit = 0; (iHit < (int) tr.fHits.size()); ++iHit) {
              const ca::Hit& h = fInputData.GetHit(tr.fHits[iHit]);
              if (fStripToTrack[h.f] == tr.fID) { fStripToTrack[h.f] = -1; }
              if (fStripToTrack[h.b] == tr.fID) { fStripToTrack[h.b] = -1; }
            }
          }
          else {
            repeatCompetition = true;
          }
        }  // itrack

      }  // competitions

      // cout << " N Competitions: " << iComp << endl;

      // ==

      for (Tindex iCandidate = 0; iCandidate < (Tindex) fTrackCandidates.size(); ++iCandidate) {
        L1Branch& tr = fTrackCandidates[iCandidate];
        if (!tr.fIsAlive) continue;
        if (kMcbm == fTrackingMode) {
          if (tr.NHits <= 3) { continue; }
        }
        else if (kGlobal == fTrackingMode) {
          if (tr.NHits < 3) { continue; }
        }
        else {
          if (tr.NHits < 3) { continue; }
        }
        if (fpCurrentIteration->GetExtendTracksFlag()) {
          if (tr.NHits != fParameters.GetNstationsActive()) BranchExtender(tr);
        }

        for (auto iHit : tr.fHits) {
          const ca::Hit& hit = fInputData.GetHit(iHit);

          /// used strips are marked

          fvHitKeyFlags[hit.f] = 1;
          fvHitKeyFlags[hit.b] = 1;

          fSliceRecoHits.push_back(iHit);
        }
        Track t;
        t.fNofHits = tr.NHits;
        fSliceRecoTracks.push_back(t);
        if (0) {  // SG debug
          cout << "store track " << iCandidate << " chi2= " << tr.chi2 << endl;
          cout << " hits: ";
          for (unsigned int ih = 0; ih < tr.fHits.size(); ih++) {
            cout << GetMcTrackIdForCaHit(tr.fHits[ih]) << " ";
          }
          cout << '\n';
        }

      }  // tracks

    }  // firstTripletLevel


    // suppress strips of suppressed hits
    for (ca::HitIndex_t ip = 0; ip < fGridPoints.size(); ip++) {
      const L1HitPoint& hp = fGridPoints[ip];
      if (hp.IsSuppressed()) {
        const ca::Hit& hit   = fGridHits[ip];
        fvHitKeyFlags[hit.f] = 1;
        fvHitKeyFlags[hit.b] = 1;
      }
    }

    if (isec < (fNFindIterations - 1)) {
      int NHitsOnStation = 0;

      for (int ista = 0; ista < fParameters.GetNstationsActive(); ++ista) {
        int start                   = fGridHitStartIndex[ista];
        int Nelements               = fGridHitStopIndex[ista] - start;
        ca::Hit* staHits              = nullptr;  // to avoid out-of-range error in ..[start]
        ca::HitIndex_t* staHitIndices = nullptr;
        L1HitPoint* staHitPoints    = nullptr;
        if (Nelements > 0) {
          staHits       = &(fGridHits[start]);
          staHitIndices = &(fGridHitIds[start]);
          staHitPoints  = &(fGridPoints[start]);
        }

        int NHitsOnStationTmp = NHitsOnStation;

        vGridTime[ista].UpdateIterGrid(Nelements, staHits, fGridHitIdsBuf, staHitIndices, fGridHitsBuf, fGridPointsBuf,
                                       staHitPoints, NHitsOnStation, *this, fvHitKeyFlags);

        fGridHitStartIndex[ista] = NHitsOnStationTmp;
        fGridHitStopIndex[ista]  = NHitsOnStation;
      }

      // #ifdef DRAW
      //     draw->ClearVeiw();
      // //   draw->DrawInfo();
      //     draw->DrawRestHits(fGridHitStartIndex, fGridHitStopIndex, RealIHit);
      //     draw->DrawRecoTracks();
      //     draw->SaveCanvas("Reco_"+isec+"_");
      //     draw->DrawAsk();
      // #endif

      // #ifdef PULLS
      //       fL1Pulls->Build(1);
      // #endif
#ifdef COUNTERS
      //  stat_nHits[isec] += (fGridHits*)->Size();

      cout << "iter = " << isec << endl;
      cout << " NSinglets = " << stat_nSinglets[isec] / stat_N << endl;
      //    cout << " NDoublets = " << stat_nDoublets[isec]/stat_N << endl;
      cout << " NTriplets = " << stat_nTriplets[isec] / stat_N << endl;
      cout << " NfGridHit = " << stat_nHits[isec] / stat_N << endl;

#endif  // COUNTERS
    }
    ++isec;  // TODO: temporary, to be removed! ()
  }          // for (int isec
  // ---- Loop over Track Finder iterations: END -----------------------------------------------------------//

  // Fit tracks
  this->L1KFTrackFitter();

  // Merge clones
  fCloneMerger.Exec(fSliceRecoTracks, fSliceRecoHits);

  //==================================

#ifdef DRAW
  draw->ClearVeiw();
  //   draw->DrawInputHits();
  //   draw->DrawInfo();
  draw->DrawRecoTracks();

  draw->SaveCanvas("Reco_");
  draw->DrawAsk();
#endif
#ifdef PULLS
  static int iEvee = 0;
  iEvee++;
  if (iEvee % 1 == 0) fL1Pulls->Build(1);
#endif
#ifdef DOUB_PERFORMANCE
  fL1Eff_doublets->CalculateEff();
  fL1Eff_doublets->Print("Doublets performance.", 1);
#endif
#ifdef TRIP_PERFORMANCE
  fL1Eff_triplets->CalculateEff();
  fL1Eff_triplets->Print("Triplet performance", 1);
//   fL1Eff_triplets2->CalculateEff();
//   fL1Eff_triplets2->Print("Triplet performance. After chi2 cut");
#endif
}


/** *************************************************************
     *                                                              *
     *     The routine performs recursive search for tracks         *
     *                                                              *
     *     I. Kisel                                    06.03.05     *
     *     I.Kulakov                                    2012        *
     *                                                              *
     ****************************************************************/

void L1Algo::CAFindTrack(int ista, L1Branch& best_tr, unsigned char& best_L, fscal& best_chi2,
                         const L1Triplet* curr_trip, L1Branch& curr_tr, unsigned char& curr_L, fscal& curr_chi2,
                         unsigned char min_best_l, L1Branch* new_tr)
/// recursive search for tracks
/// input: @ista - station index, @&best_tr - best track for the privious call, @&best_L -
/// output: @&NCalls - number of function calls
{
  if (curr_trip->GetLevel() == 0)  // the end of the track -> check and store
  {
    // -- finish with current track
    // add rest of hits
    const ca::HitIndex_t& ihitm = curr_trip->GetMHit();
    const ca::HitIndex_t& ihitr = curr_trip->GetRHit();


    //if (!GetFUsed((*fStripFlag)[fGridHits[ihitm].f] | (*fStripFlag)[fGridHits[ihitm].b])) {
    //L1_SHOW(fInputData.GetNhitKeys());
    //L1_SHOW(fvHitKeyFlags.size());
    //L1_SHOW(fGridHits[ihitm].f);
    //L1_SHOW(fGridHits[ihitm].b);
    if (!(fvHitKeyFlags[fGridHits[ihitm].f] || fvHitKeyFlags[fGridHits[ihitm].b])) {

      //        curr_tr.Hits.push_back(fGridHitIds[ihitm]);

      curr_tr.fHits.push_back(fGridHitIds[ihitm]);

      curr_tr.NHits++;

      curr_L++;
    }

    //if (!GetFUsed((*fStripFlag)[fGridHits[ihitr].f] | (*fStripFlag)[fGridHits[ihitr].b])) {
    //L1_SHOW(fInputData.GetNhitKeys());
    //L1_SHOW(fvHitKeyFlags.size());
    //L1_SHOW(fGridHits[ihitr].f);
    //L1_SHOW(fGridHits[ihitr].b);
    if (!(fvHitKeyFlags[fGridHits[ihitr].f] || fvHitKeyFlags[fGridHits[ihitr].b])) {

      //curr_tr.Hits.push_back(fGridHitIds[ihitr]);
      curr_tr.fHits.push_back(fGridHitIds[ihitr]);

      curr_tr.NHits++;

      curr_L++;
    }

    //if( curr_L < min_best_l - 1 ) return; // suppouse that only one hit can be added by extender
    // TODO: SZh 21.08.2023: Replace hard-coded value with a parameter
    int ndf = curr_L * 2 - 5;

    if (kGlobal == fTrackingMode || kMcbm == fTrackingMode) { ndf = curr_L * 2 - 4; }

    if (curr_chi2 > fTrackChi2Cut * ndf) { return; }

    // TODO: SZh 04.10.2022: Does one should delete this code?
    //       // try to find more hits
    // if (fpCurrentIteration->GetExtendTracksFlag()) {
    //     // if( curr_L < min_best_l )
    //     if (!fpCurrentIteration->GetJumpedFlag() && curr_L >= 3){
    //       //curr_chi2 = BranchExtender(curr_tr);
    //       BranchExtender(curr_tr);
    //       curr_L = curr_tr.Hits.size();
    //         //      if( 2*curr_chi2 > (2*(curr_L*2-5) + 1) * 4*4 ) return;
    //     }
    // }

    // -- select the best
    if ((curr_L > best_L) || ((curr_L == best_L) && (curr_chi2 < best_chi2))) {

      best_tr   = curr_tr;
      best_chi2 = curr_chi2;
      best_L    = curr_L;
    }

    return;
  }
  else  //MEANS level ! = 0
  {
    int N_neighbour = (curr_trip->GetNNeighbours());

    for (Tindex in = 0; in < N_neighbour; in++) {

      unsigned int ID = curr_trip->GetFNeighbour() + in;

      unsigned int Station = TripletId2Station(ID);
      unsigned int Triplet = TripletId2Triplet(ID);

      const L1Triplet& new_trip = fTriplets[Station][Triplet];

      fscal dchi2 = 0.;
      if (!checkTripletMatch(*curr_trip, new_trip, dchi2)) continue;

      if (fvHitKeyFlags[fGridHits[new_trip.GetLHit()].f]
          || fvHitKeyFlags[fGridHits[new_trip.GetLHit()].b]) {  //hits are used
        //  no used hits allowed -> compare and store track
        if ((curr_L > best_L) || ((curr_L == best_L) && (curr_chi2 < best_chi2))) {
          best_tr = curr_tr;

          best_chi2 = curr_chi2;
          best_L    = curr_L;
        }
      }
      else {  // hit is not used: add the left hit from the new triplet to the current track

        unsigned char new_L = curr_L + 1;
        fscal new_chi2      = curr_chi2 + dchi2;


        if (0) {  //SGtrd2d debug!!
          int mc01 = GetMcTrackIdForGridHit(curr_trip->GetLHit());
          int mc02 = GetMcTrackIdForGridHit(curr_trip->GetMHit());
          int mc03 = GetMcTrackIdForGridHit(curr_trip->GetRHit());
          int mc11 = GetMcTrackIdForGridHit(new_trip.GetLHit());
          int mc12 = GetMcTrackIdForGridHit(new_trip.GetMHit());
          int mc13 = GetMcTrackIdForGridHit(new_trip.GetRHit());
          if ((mc01 == mc02) && (mc02 == mc03)) {
            cout << " sta " << ista << " mc0 " << mc01 << " " << mc02 << " " << mc03 << " mc1 " << mc11 << " " << mc12
                 << " " << mc13 << " chi2 " << curr_chi2 / (2 * (curr_L + 2) - 4) << " new "
                 << new_chi2 / (2 * (new_L + 2) - 4) << endl;
            cout << "   hits " << curr_trip->GetLHit() << " " << curr_trip->GetMHit() << " " << curr_trip->GetRHit()
                 << " " << new_trip.GetLHit() << endl;
          }
        }

        int ndf = 2 * (new_L + 2) - 5;

        if (kGlobal == fTrackingMode) {  //SGtrd2d!!!
          ndf = 2 * (new_L + 2) - 4;
        }
        else if (kMcbm == fTrackingMode) {
          ndf = 2 * (new_L + 2) - 4;
        }
        else {
          ndf = new_L;  // TODO: SG:  2 * (new_L + 2) - 5
        }

        if (new_chi2 > fTrackChi2Cut * ndf) continue;

        // add new hit
        new_tr[ista] = curr_tr;
        new_tr[ista].fHits.push_back(fGridHitIds[new_trip.GetLHit()]);
        new_tr[ista].NHits++;

        const int new_ista = ista + new_trip.GetMSta() - new_trip.GetLSta();

        CAFindTrack(new_ista, best_tr, best_L, best_chi2, &new_trip, new_tr[ista], new_L, new_chi2, min_best_l, new_tr);
      }  // add triplet to track
    }    // for neighbours
  }      // level = 0
}

#ifdef DRAW
void L1Algo::DrawRecoTracksTime(const cbm::algo::ca::Vector<CbmL1Track>& tracks)
{
  // TODO: find where the DrawRecoTracksTime is. It is missing in the git repository.
  //draw->DrawRecoTracksTime(tracks);
  draw->SaveCanvas(" ");
}
#endif
