/* Copyright (C) 2021-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/// \file  L1InitManager.h
/// \brief Input data management class for L1Algo
/// \since 24.12.2021
#ifndef L1InitManager_h
#define L1InitManager_h 1

#include <array>
#include <bitset>
#include <memory>  //unique_ptr
#include <numeric>
#include <set>
#include <string>
#include <type_traits>
#include <unordered_map>

#include "CaConstants.h"
#include "CaField.h"
#include "CaIteration.h"
#include "CaObjectInitController.h"
#include "CaParameters.h"
#include "CaSimd.h"
#include "L1BaseStationInfo.h"
#include "L1Utils.h"


using namespace cbm::algo::ca;  //TODO: remove
using namespace cbm::algo;      //TODO: remove

class L1ConfigRW;
class L1Algo;

/// Forward declaration of the tracking detectors scoped enumeration. Concrete realization of this enumeration must be
/// determined in the concrete setup class (i.e. CbmL1/BmnL1)
enum class L1DetectorID;

/// Underlying
using L1DetectorID_t = std::underlying_type_t<L1DetectorID>;

/// Initialization manager for L1Algo
///
/// ==== Expected initialization steps ==== (TODO: keep this instruction up-to-date)
///
/// 0. Implement your enum class L1DetectorID with detector systems used for tracking:
///    IMPORTANT!!! Detectors must be sorted in the beam direction
///    /// Concrete L1DetectorID implementation for CBM
///    enum class L1DetectorID {
///      kMvd,
///      kSts,
///      kMuch,
///      kTof,
///      kTrd
///    };
///    For CBM see implementation in CbmL1DetectorID.h
///
/// 1. Get a pointer to the :L1InitManager field of L1Algo:
///
///    L1InitManager * initMan = algo->GetL1InitManager();
///
/// 2. Initialize a set of L1DetectorID's for detectors active in tracking:
///
///    std::set<L1DetectorID> activeTrackingDetectors { L1Detector::kMvd, L1Detector::kSts };
///    initMan->SetActiveDetectorIDs(activeTrackingDetectors);
///
/// 3. Initialize number of stations for each detector:
///
///    initMan->SetNstationsCrosscheck(L1DetectorID::kMvd, fNMvdStations)
///    initMan->SetNstationsCrosscheck(L1DetectorID::kSts, fNStsStations);
///
/// 4. Initialize each station using L1BaseStationInfo:
///
// TODO: Implement mechanism of reinitialization (S.Zharko)
class L1InitManager {
private:
  enum class EInitKey
  {
    // NOTE: Please, keep the numbers of the enumeration items in the existing order: it helps to debug the
    //       initialization with this->GetObjectInitController().ToString() method call (S.Zharko)
    kActiveDetectorIDs,             ///< 0) If the detector sequence is set
    kFieldFunction,                 ///< 1) If magnetic field getter function is set
    kTargetPos,                     ///< 2) If target position was defined
    kPrimaryVertexField,            ///< 3) If magnetic field value and region defined at primary vertex
    kStationsInfo,                  ///< 4) If all the planned stations were added to the manager
    kCAIterationsNumberCrosscheck,  ///< 5) If the number of CA track finder is initialized
    kCAIterations,                  ///< 6) If the CA track finder iterations were initialized
    kSearchWindows,                 ///< 7) If the hit search windows were initialized
    kGhostSuppression,              ///< 9)
    kRandomSeed,                    ///< 10) If the random seed is provided
    kStationLayoutInitialized,      ///< 11) If stations layout is initialized
    kEnd                            ///< 12) [technical] number of entries in the enumeration
  };

  using L1DetectorIDIntMap_t = std::unordered_map<L1DetectorID, int, L1Utils::EnumClassHash>;
  using L1DetectorIDSet_t    = std::set<L1DetectorID>;
  using L1FieldFunction_t    = std::function<void(const double (&xyz)[3], double (&B)[3])>;
  using InitController_t     = ObjectInitController<static_cast<int>(EInitKey::kEnd), EInitKey>;
  template<typename T>
  using L1DetectorIDArr_t = std::array<T, constants::size::MaxNdetectors>;

public:
  /// Default constructor
  L1InitManager() = default;

  /// Destructor
  ~L1InitManager() = default;

  /// Copy constructor is forbidden
  L1InitManager(const L1InitManager& /*other*/) = delete;

  /// Move constructor is forbidden
  L1InitManager(L1InitManager&& /*other*/) = delete;

  /// Copy assignment operator is forbidden
  L1InitManager& operator=(const L1InitManager& /*other*/) = delete;

  /// Move assignment operator is forbidden
  L1InitManager& operator=(L1InitManager&& /*other*/) = delete;


  /// @brief Adds a tracking station to the geometry
  ///
  /// @note The added station stays uninitialized until the Parameters object is formed
  void AddStation(const L1BaseStationInfo& station);

  /// Provides final checks of large fields initialization calling Check"Object"Init() private methods,
  /// must be called in the beginning of L1Algo::Init()
  void CheckInit();
  // NOTE: This method calls checkers of large fields initializations like a station or an iteration. The method must be
  //       called in the L1Algo class. (S.Zharko)

  /// Clears vector of CA track finder iterations
  void ClearCAIterations();

  /// Clears vector of base setup
  void ClearSetupInfo();

  /// Forms parameters container
  /// \return  Success flag
  bool FormParametersContainer();

  /// @brief  Gets name of the detector
  /// @param  detId Index of the detector
  /// @return Name of the detector
  const std::string& GetDetectorName(L1DetectorID detId) const { return fvDetectorNames[static_cast<int>(detId)]; }

  /// Gets ghost suppression flag
  int GetGhostSuppression() const { return fParameters.fGhostSuppression; }

  /// @brief Gets a name of the main input configuration file
  const std::string& GetInputConfigMain() const { return fsConfigInputMain; }

  /// @brief Gets a name of the user input configuration file
  const std::string& GetInputConfigUser() const { return fsConfigInputMain; }

  /// Gets a const reference to ca::ObjectInitController
  const InitController_t& GetInitController() const { return fInitController; }

  /// Gets total number of active stations
  int GetNstationsActive() const;

  /// Gets number of active stations for given detector ID
  int GetNstationsActive(L1DetectorID detectorID) const;

  /// Gets total number of stations, provided by setup geometry
  int GetNstationsGeometry() const;

  /// Gets number of stations, provided by setup geometry for given detector ID
  int GetNstationsGeometry(L1DetectorID detectorID) const;

  /// Gets a name of the output configuration file
  const std::string& GetOutputConfigName() const { return fConfigOutputName; }

  /// @brief Gets a reference to the stations array
  std::vector<L1BaseStationInfo>& GetStationInfo();

  /// @brief Initializes station layout
  ///
  /// This function is to be called after all the tracking stations (L1BaseStationInfo objects) are added to the
  /// L1InitManager instance. After the initialization a vector of the tracking stations sorted by z-positions is
  /// available for modifications.
  void InitStationLayout();

  /// @brief Calculates ca::FieldValue and L1FieldReference values for a selected step in z-axis from the target position
  /// \param zStep step between nodal points
  // TODO: Consider possibility for linear approximation (S.Zh.)
  void InitTargetField(double zStep);

  /// Pushes an CA track finder iteration into a sequence of iteration using reference
  void PushBackCAIteration(const Iteration& iteration);

  /// Pushes an CA track finder iteration into a sequence of iteration using raw pointer
  void PushBackCAIteration(const Iteration* pIteration) { PushBackCAIteration(*pIteration); }

  /// Pushes an CA track finder iteration into a sequence of iteration using std::unique_ptr
  void PushBackCAIteration(const std::unique_ptr<Iteration>& puIteration) { PushBackCAIteration(*puIteration); }

  /// @brief Reads main and user parameters configs
  void ReadInputConfigs();

  /// Reads parameters object from boost-serialized binary file
  /// \param  fileName  Name of input file
  void ReadParametersObject(const std::string& fileName);

  /// Reads search windows from file
  /// \param  fileName  Name of input file
  void ReadSearchWindows(const std::string& fileName);

  /// Sets a set of active tracking detector IDs
  void SetActiveDetectorIDs(const L1DetectorIDSet_t& detectorIDs);

  /// Sets a number of CA track finder iterations to provide initialization cross-check
  void SetCAIterationsNumberCrosscheck(int nIterations);

  /// @brief Sets base configuration file
  /// @param mainConfig  Path to main configuration file
  /// @note  The base configuraiton file is mandatory until the tracking configuration is initialized from
  ///        beforehand created Parameters file.
  void SetConfigMain(const std::string& mainConfig) { fsConfigInputMain = mainConfig; }

  /// @brief Sets user configuration file
  /// @param userConfig  Path to user configuration file
  /// @note  The user configuraiton file is optional
  void SetConfigUser(const std::string& userConfig) { fsConfigInputUser = userConfig; }

  /// @brief Sets detector names
  /// @param container  Container of the detector names
  template<size_t Size>
  void SetDetectorNames(const std::array<const char*, Size>& container)
  {
    static_assert(Size <= constants::size::MaxNdetectors,
                  "Please, be ensured that the constants::size::MaxNdetectors is not lower then the "
                  "L1DetectorID::kEND value, provided by your setup");
    std::copy(container.begin(), container.end(), fvDetectorNames.begin());
  }

  /// Sets a magnetic field function, which will be applied for all the stations
  void SetFieldFunction(const L1FieldFunction_t& fieldFcn);

  /// FIXME: ...
  void SetGhostSuppression(int ghostSuppression);

  /// Sets a name of the output configuration file. The output file is created from the fields, saved in the resulted
  /// Parameters object
  /// \param  filename  Name of the output L1 parameters configuration
  void SetOutputConfigName(const std::string& filename) { fConfigOutputName = filename; }

  /// @brief Sets pseudo-random numbers generator seed
  /// @param seed  Seed value
  /// @note  The default seed is 1
  void SetRandomSeed(unsigned int seed);

  /// Sets target position
  /// \param  x  Position X component [cm]
  /// \param  y  Position Y component [cm]
  /// \param  z  Position Z component [cm]
  void SetTargetPosition(double x, double y, double z);

  /// Sets upper-bound cut on max number of doublets per one singlet
  void SetMaxDoubletsPerSinglet(unsigned int value) { fParameters.fMaxDoubletsPerSinglet = value; }

  /// Sets upper-bound cut on max number of triplets per one doublet
  void SetMaxTripletPerDoublets(unsigned int value) { fParameters.fMaxTripletPerDoublets = value; }

  /// Sends (moves) parameters object to the destination reference
  /// \param  pAlgo  Pointer to the L1 tracking algorithm main class
  /// \return Success flag
  bool SendParameters(L1Algo* pAlgo);

  /// @brief Sends (moves) parameters object to an external destination
  /// @param destination  Reference to the parameters object of destination
  /// @return  Success flag
  bool SendParameters(Parameters& destination);

  /// Writes parameters object from boost-serialized binary file
  /// \param  fileName  Name of input file
  void WriteParametersObject(const std::string& fileName) const;


  // ***************************
  // ** Flags for development **
  // ***************************

  /// Ignore hit search areas
  void DevSetIgnoreHitSearchAreas(bool value = true) { fParameters.fDevIsIgnoreHitSearchAreas = value; }

  /// Force use of the original field (not approximated)
  void DevSetUseOfOriginalField(bool value = true) { fParameters.fDevIsUseOfOriginalField = value; }

  /// Flag to match doublets using MC information
  void DevSetIsMatchDoubletsViaMc(bool value = true) { fParameters.fDevIsMatchDoubletsViaMc = value; }

  /// Flag to match triplets using Mc information
  void DevSetIsMatchTripletsViaMc(bool value = true) { fParameters.fDevIsMatchTripletsViaMc = value; }

  /// Flag to match triplets using Mc information
  void DevSetIsExtendTracksViaMc(bool value = true) { fParameters.fDevIsExtendTracksViaMc = value; }

  /// Flag to match triplets using Mc information
  void DevSetIsSuppressOverlapHitsViaMc(bool value = true) { fParameters.fDevIsSuppressOverlapHitsViaMc = value; }

  /// Flag to use estimated hit search windows
  /// true:  estimated search windows will be used in track finder
  /// false: Kalman filter will be used in track finder
  void DevSetIsParSearchWUsed(bool value = true) { fParameters.fDevIsParSearchWUsed = value; }

private:
  // *********************
  // ** Private methods **
  // *********************

  /// Checker for L1CAIteration container initialization (sets EInitKey::kCAIterations)
  /// \return true If all L1CAIteration objects were initialized properly
  void CheckCAIterationsInit();

  /// Checker for L1BaseStationInfo set initialization (sets EInitKey::kStationsInfo)
  /// \return true If all L1BaseStationInfo objects were initialized properly. Similar effect can be achieved by
  void CheckStationsInfoInit();

  /// @brief Returns station layout into undefined condition
  void ClearStationLayout();


  // *****************
  // ** Data fields **
  // *****************

  // * Basic fields *

  InitController_t fInitController {};                ///< Initialization flags
  L1DetectorIDSet_t fActiveDetectorIDs {};            ///< Set of tracking detectors, active during the run
  L1DetectorIDArr_t<std::string> fvDetectorNames {};  ///< Names of the detectors

  // * Target fields *

  double fTargetZ {0.};  ///< Target position z component in double precision

  // * Stations related fields *

  std::vector<L1BaseStationInfo> fvStationInfo {};  ///< Vector of L1BaseStationInfo objects (active + inactive)

  /// A function which returns magnetic field vector B in a radius-vector xyz
  L1FieldFunction_t fFieldFunction {[](const double (&)[3], double (&)[3]) {}};
  // NOTE: Stations of detectors which will not be assigned as active, will not be included in the tracking!

  // * CA track finder iterations related *

  int fCAIterationsNumberCrosscheck {-1};  ///< Number of iterations to be passed (must be used for cross-checks)

  Parameters fParameters {};  ///< L1Algo parameters object

  // * Configuration related *
  std::string fsConfigInputMain = "";  ///< name for the input configuration file
  std::string fsConfigInputUser = "";  ///< name for the input configuration file
  std::string fConfigOutputName = "";  ///< name for the output configuration file
  //L1ConfigRW fConfigRW {this, /*verbose = */ 4};  ///< configuration file reader and writer
};

#endif
