/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

#include "L1MaterialMonitor.h"

#include <Logger.h>

#include <iomanip>
#include <sstream>


void L1MaterialMonitor::SetMaterial(const ca::MaterialMap* materialMap)
{
  /// construction
  /// \param  materialMap  Matareial map to be monitored

  fMaterial = materialMap;

  fActiveBinMap.resize(0);

  if (fMaterial) { fActiveBinMap.assign(fMaterial->GetNbins() * fMaterial->GetNbins(), 0); }

  fNhitsTotal   = 0;
  fNhitsOutside = 0;

  EvaluateStatistics();
}

void L1MaterialMonitor::MarkActiveBin(float x, float y)
{
  /// mark a bin as active
  if (!fMaterial) {
    LOG(fatal) << "L1MaterialMonitor: material map is not set";
    return;
  }
  int i = fMaterial->GetBin(x, y);
  fNhitsTotal++;
  if (i < 0) { fNhitsOutside++; }
  else {
    fActiveBinMap[i] = 1;
  }
}

void L1MaterialMonitor::EvaluateStatistics()
{
  /// update values of statistical variables with respect to the active map

  fActiveNbins  = 0;
  fPassiveNbins = 0;

  fActiveRadThickMin  = constants::Undef<double>;
  fActiveRadThickMax  = constants::Undef<double>;
  fActiveRadThickMean = constants::Undef<double>;

  fPassiveRadThickMin  = constants::Undef<double>;
  fPassiveRadThickMax  = constants::Undef<double>;
  fPassiveRadThickMean = constants::Undef<double>;

  if (!fMaterial) { return; }

  int nBins = fMaterial->GetNbins() * fMaterial->GetNbins();

  if (nBins != (int) fActiveBinMap.size()) {
    LOG(fatal) << "L1MaterialMonitor: map of active bins is not consistent with the mterial map: nbins "
               << fActiveBinMap.size() << " != " << nBins;
    return;
  }

  for (int i = 0; i < nBins; i++) {
    double r = fMaterial->GetRadThickBin(i);
    if (fActiveBinMap[i]) {  // active material
      if (fActiveNbins == 0) {
        fActiveRadThickMin  = r;
        fActiveRadThickMax  = r;
        fActiveRadThickMean = r;
      }
      else {
        fActiveRadThickMin = std::min(fActiveRadThickMin, r);
        fActiveRadThickMax = std::max(fActiveRadThickMax, r);
        fActiveRadThickMean += r;
      }
      fActiveNbins++;
    }
    else {
      // passive material
      if (fPassiveNbins == 0) {
        fPassiveRadThickMin  = r;
        fPassiveRadThickMax  = r;
        fPassiveRadThickMean = r;
      }
      else {
        fPassiveRadThickMin = std::min(fPassiveRadThickMin, r);
        fPassiveRadThickMax = std::max(fPassiveRadThickMax, r);
        fPassiveRadThickMean += r;
      }
      fPassiveNbins++;
    }
  }
  if (fActiveNbins + fPassiveNbins != nBins) {
    LOG(fatal) << "L1MaterialMonitor: wrong calculation of N passive / active bins ";
  }
  if (fActiveNbins > 0) { fActiveRadThickMean /= fActiveNbins; }
  if (fPassiveNbins > 0) { fPassiveRadThickMean /= fPassiveNbins; }
}


TString L1MaterialMonitor::ToString()
{
  /// print statistics to a string

  EvaluateStatistics();

  TString s(Form("material map %s. ", fName.Data()));

  if (fActiveNbins > 0) {
    s += Form("Active material RL: min %.2f%%, max %.2f%%, mean %.2f%%. ", fActiveRadThickMin * 100.,
              fActiveRadThickMax * 100., fActiveRadThickMean * 100);
  }
  else {
    if (fNhitsTotal > 0) { s += "No active material. "; }
    else {
      s += "No hits to identify active areas. ";
    }
  }

  if (fPassiveNbins > 0) {
    s += Form("Passive material RL: min %.2f%%, max %.2f%%, mean %.2f%%. ", fPassiveRadThickMin * 100.,
              fPassiveRadThickMax * 100., fPassiveRadThickMean * 100);
  }
  else {
    s += "No passive material. ";
  }

  if (fNhitsTotal > 0) {
    if (fNhitsOutside == 0) { s += "No hits outside of the map. "; }
    else {
      s += Form("There are %f%% hits outside the map!!! ", (100. * fNhitsOutside) / fNhitsTotal);
    }
  }
  else {
    s += "No hit statistics. ";
  }

  return s;
}
