/* Copyright (C) 2010-2020 Frankfurt Institute for Advanced Studies, Goethe-Universitaet Frankfurt, Frankfurt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Ivan Kisel, Sergey Gorbunov, Igor Kulakov [committer], Valentina Akishina, Maksym Zyzak */

#include <iostream>
#include <vector>

#include "CaTrackParam.h"
#include "L1Algo.h"
#include "L1Fit.h"

using std::cout;
using std::endl;
using std::vector;
using Track = cbm::algo::ca::Track;

namespace NS_L1TrackFitter
{
  const fvec c_light(0.000299792458), c_light_i(fvec(1.) / c_light);
}  // namespace NS_L1TrackFitter
using namespace NS_L1TrackFitter;


void L1Algo::L1KFTrackFitter()
{
  //  cout << " Start L1 Track Fitter " << endl;
  int start_hit = 0;  // for interation in fSliceRecoHits[]

  //  static ca::FieldValue fldB0, fldB1, fldB2 _fvecalignment;
  //  static ca::FieldRegion fld _fvecalignment;
  ca::FieldValue fldB0, fldB1, fldB2 _fvecalignment;
  ca::FieldRegion fld _fvecalignment;


  ca::FieldValue fldB01, fldB11, fldB21 _fvecalignment;
  ca::FieldRegion fld1 _fvecalignment;

  const int nStations = fParameters.GetNstationsActive();
  int nTracks_SIMD    = fvec::size();

  L1Fit fit;  // fit parameters coresponding to the current track
  TrackParamV& tr = fit.Tr();
  fit.SetParticleMass(GetDefaultParticleMass());
  fit.SetDoFitVelocity(true);

  Track* t[fvec::size()] {nullptr};

  const ca::Station* sta = fParameters.GetStations().begin();

  // Spatial-time position of a hit vs. station and track in the portion

  fvec x[constants::size::MaxNstations];                      // Hit position along the x-axis [cm]
  fvec y[constants::size::MaxNstations];                      // Hit position along the y-axis [cm]
  L1XYMeasurementInfo cov_xy[constants::size::MaxNstations];  // Covariance matrix for x,y

  fvec z[constants::size::MaxNstations];  // Hit position along the z-axis (precised) [cm]

  fvec time[constants::size::MaxNstations];  // Hit time [ns]
  fvec dt2[constants::size::MaxNstations];   // Hit time uncertainty [ns] squared

  fvec x_first;
  fvec y_first;
  L1XYMeasurementInfo cov_xy_fst;

  fvec time_first;
  fvec wtime_first;
  fvec dt2_first;

  fvec x_last;
  fvec y_last;
  L1XYMeasurementInfo cov_xy_lst;

  fvec time_last;
  fvec wtime_last;
  fvec dt2_last;

  fvec By[constants::size::MaxNstations];
  fmask w[constants::size::MaxNstations];
  fmask w_time[constants::size::MaxNstations];  // !!!

  fvec y_temp;
  fvec x_temp;
  fvec fldZ0;
  fvec fldZ1;
  fvec fldZ2;
  fvec z_start;
  fvec z_end;

  ca::FieldValue fB[constants::size::MaxNstations], fB_temp _fvecalignment;


  fvec ZSta[constants::size::MaxNstations];
  for (int ista = 0; ista < nStations; ista++) {
    ZSta[ista]       = sta[ista].fZ;
    cov_xy[ista].C00 = 1.f;
    cov_xy[ista].C10 = 0.f;
    cov_xy[ista].C11 = 1.f;
  }

  unsigned short N_vTracks = fSliceRecoTracks.size();  // number of tracks processed per one SSE register

  for (unsigned short itrack = 0; itrack < N_vTracks; itrack += fvec::size()) {

    if (N_vTracks - itrack < static_cast<unsigned short>(fvec::size())) nTracks_SIMD = N_vTracks - itrack;

    for (int i = 0; i < nTracks_SIMD; i++) {
      t[i] = &fSliceRecoTracks[itrack + i];
    }

    // fill the rest of the SIMD vectors with something reasonable
    for (uint i = nTracks_SIMD; i < fvec::size(); i++) {
      t[i] = &fSliceRecoTracks[itrack];
    }

    // get hits of current track
    for (int ista = 0; ista < nStations; ista++) {
      w[ista]      = fmask::Zero();
      w_time[ista] = fmask::Zero();
      z[ista]      = ZSta[ista];
    }

    //fmask isFieldPresent = fmask::Zero();

    for (int iVec = 0; iVec < nTracks_SIMD; iVec++) {

      int nHitsTrack = t[iVec]->fNofHits;
      int iSta[constants::size::MaxNstations];

      for (int ih = 0; ih < nHitsTrack; ih++) {

        const ca::Hit& hit = fInputData.GetHit(fSliceRecoHits[start_hit++]);
        const int ista   = hit.iSt;

        //if (sta[ista].fieldStatus) { isFieldPresent[iVec] = true; }

        iSta[ih]      = ista;
        w[ista][iVec] = true;
        if (sta[ista].timeInfo) { w_time[ista][iVec] = true; }

        x[ista][iVec]    = hit.x;  //x_temp[iVec];
        y[ista][iVec]    = hit.y;  //y_temp[iVec];
        time[ista][iVec] = hit.t;
        dt2[ista][iVec]  = hit.dt2;
        if (!sta[ista].timeInfo) { dt2[ista][iVec] = 1.e4; }
        z[ista][iVec] = hit.z;
        fB_temp                = sta[ista].fieldSlice.GetFieldValue(x[ista], y[ista]);
        cov_xy[ista].C00[iVec] = hit.dx2;
        cov_xy[ista].C10[iVec] = hit.dxy;
        cov_xy[ista].C11[iVec] = hit.dy2;

        fB[ista].x[iVec] = fB_temp.x[iVec];
        fB[ista].y[iVec] = fB_temp.y[iVec];
        fB[ista].z[iVec] = fB_temp.z[iVec];
        if (ih == 0) {
          z_start[iVec]        = z[ista][iVec];
          x_first[iVec]        = x[ista][iVec];
          y_first[iVec]        = y[ista][iVec];
          time_first[iVec]     = time[ista][iVec];
          wtime_first[iVec]    = sta[ista].timeInfo ? 1. : 0.;
          dt2_first[iVec]      = dt2[ista][iVec];
          cov_xy_fst.C00[iVec] = cov_xy[ista].C00[iVec];
          cov_xy_fst.C10[iVec] = cov_xy[ista].C10[iVec];
          cov_xy_fst.C11[iVec] = cov_xy[ista].C11[iVec];
        }
        else if (ih == nHitsTrack - 1) {
          z_end[iVec]          = z[ista][iVec];
          x_last[iVec]         = x[ista][iVec];
          y_last[iVec]         = y[ista][iVec];
          cov_xy_lst.C00[iVec] = cov_xy[ista].C00[iVec];
          cov_xy_lst.C10[iVec] = cov_xy[ista].C10[iVec];
          cov_xy_lst.C11[iVec] = cov_xy[ista].C11[iVec];
          time_last[iVec]      = time[ista][iVec];
          dt2_last[iVec]       = dt2[ista][iVec];
          wtime_last[iVec]     = sta[ista].timeInfo ? 1. : 0.;
        }
      }

      for (int ih = nHitsTrack - 1; ih >= 0; ih--) {
        const int ista      = iSta[ih];
        const ca::Station& st = fParameters.GetStation(ista);
        By[ista][iVec]      = st.fieldSlice.cy[0][0];
      }
    }

    fit.GuessTrack(z_end, x, y, z, time, By, w, w_time, nStations);

    if (kGlobal == fTrackingMode || kMcbm == fTrackingMode) { tr.Qp() = fvec(1. / 1.1); }

    // tr.Qp() = iif(isFieldPresent, tr.Qp(), fvec(1. / 0.25));

    for (int iter = 0; iter < 2; iter++) {  // 1.5 iterations

      fit.SetQp0(tr.Qp());

      // fit backward

      int ista = nStations - 1;

      time_last = iif(w_time[ista], time_last, fvec::Zero());
      dt2_last  = iif(w_time[ista], dt2_last, fvec(1.e6));

      tr.ResetErrors(cov_xy_lst.C00, cov_xy_lst.C11, 0.1, 0.1, 1.0, dt2_last, 1.e-2);
      tr.C10()  = cov_xy_lst.C10;
      tr.X()    = x_last;
      tr.Y()    = y_last;
      tr.Time() = time_last;
      tr.Vi()   = constants::phys::SpeedOfLightInv;
      tr.InitVelocityRange(0.5);
      tr.Ndf()     = fvec(-5.) + fvec(2.);
      tr.NdfTime() = fvec(-2.) + wtime_last;

      fldZ1 = z[ista];

      fldB1 = sta[ista].fieldSlice.GetFieldValue(tr.X(), tr.Y());

      fldB1.Combine(fB[ista], w[ista]);

      fldZ2   = z[ista - 2];
      fvec dz = fldZ2 - fldZ1;
      fldB2   = sta[ista].fieldSlice.GetFieldValue(tr.X() + tr.Tx() * dz, tr.Y() + tr.Ty() * dz);
      fldB2.Combine(fB[ista - 2], w[ista - 2]);
      fld.Set(fldB2, fldZ2, fldB1, fldZ1, fldB0, fldZ0);

      for (--ista; ista >= 0; ista--) {

        fldZ0 = z[ista];
        dz    = (fldZ1 - fldZ0);
        fldB0 = sta[ista].fieldSlice.GetFieldValue(tr.X() - tr.Tx() * dz, tr.Y() - tr.Ty() * dz);
        fldB0.Combine(fB[ista], w[ista]);
        fld.Set(fldB0, fldZ0, fldB1, fldZ1, fldB2, fldZ2);

        fmask initialised = (z[ista] < z_end) & (z_start <= z[ista]);

        fld1 = fld;

        fit.SetMask(initialised);
        fit.Extrapolate(z[ista], fld1);
        fit.AddMsInMaterial(fParameters.GetMaterialThickness(ista, tr.X(), tr.Y()));
        fit.EnergyLossCorrection(fParameters.GetMaterialThickness(ista, tr.X(), tr.Y()), fvec(1.f));

        fit.SetMask(initialised && w[ista]);
        fit.FilterXY(cov_xy[ista], x[ista], y[ista]);
        fit.FilterTime(time[ista], dt2[ista], sta[ista].timeInfo);


        fldB2 = fldB1;
        fldZ2 = fldZ1;
        fldB1 = fldB0;
        fldZ1 = fldZ0;
      }

      // extrapolate to the PV region

      L1Fit fitpv = fit;
      {
        fitpv.SetMask(fmask::One());

        L1XYMeasurementInfo vtxInfo;
        vtxInfo.C00 = fvec(1.e-8);
        vtxInfo.C10 = fvec::Zero();
        vtxInfo.C11 = fvec(1.e-8);

        if (kGlobal == fTrackingMode) {
          for (int vtxIter = 0; vtxIter < 2; vtxIter++) {
            fitpv.SetQp0(fitpv.Tr().Qp());
            fitpv.Tr()    = fit.Tr();
            fitpv.Tr().Qp() = fitpv.Qp0();
            fitpv.Extrapolate(fParameters.GetTargetPositionZ(), fld);
            fitpv.FilterXY(vtxInfo, fParameters.GetTargetPositionX(), fParameters.GetTargetPositionY());
          }
        }
        else {
          fitpv.SetQp0(fitpv.Tr().Qp());
          fitpv.Extrapolate(fParameters.GetTargetPositionZ(), fld);
        }
      }

      //fit.SetQp0(tr.Qp());
      //fit.SetMask(fmask::One());
      //fit.MeasureVelocityWithQp();
      //fit.FilterVi(TrackParamV::kClightNsInv);

      TrackParamV Tf = fit.Tr();
      if (kGlobal == fTrackingMode) { Tf.Qp() = fitpv.Tr().Qp(); }

      for (int iVec = 0; iVec < nTracks_SIMD; iVec++) {
        t[iVec]->fParFirst.Set(Tf, iVec);
      }

      for (int iVec = 0; iVec < nTracks_SIMD; iVec++) {
        t[iVec]->fParPV.Set(fitpv.Tr(), iVec);
      }

      if (iter == 1) { break; }  // only 1.5 iterations

      // fit forward

      ista = 0;

      tr.ResetErrors(cov_xy_fst.C00, cov_xy_fst.C11, 0.1, 0.1, 1., dt2_first, 1.e-2);
      tr.C10() = cov_xy_fst.C10;

      tr.X()    = x_first;
      tr.Y()    = y_first;
      tr.Time() = time_first;
      tr.Vi()   = constants::phys::SpeedOfLightInv;
      tr.InitVelocityRange(0.5);

      tr.Ndf()     = fvec(-5. + 2.);
      tr.NdfTime() = fvec(-2.) + wtime_first;

      fit.SetQp0(tr.Qp());

      fldZ1 = z[ista];
      fldB1 = sta[ista].fieldSlice.GetFieldValue(tr.X(), tr.Y());
      fldB1.Combine(fB[ista], w[ista]);

      fldZ2 = z[ista + 2];
      dz    = fldZ2 - fldZ1;
      fldB2 = sta[ista].fieldSlice.GetFieldValue(tr.X() + tr.Tx() * dz, tr.Y() + tr.Ty() * dz);
      fldB2.Combine(fB[ista + 2], w[ista + 2]);
      fld.Set(fldB2, fldZ2, fldB1, fldZ1, fldB0, fldZ0);

      for (++ista; ista < nStations; ista++) {
        fldZ0 = z[ista];
        dz    = (fldZ1 - fldZ0);
        fldB0 = sta[ista].fieldSlice.GetFieldValue(tr.X() - tr.Tx() * dz, tr.Y() - tr.Ty() * dz);
        fldB0.Combine(fB[ista], w[ista]);
        fld.Set(fldB0, fldZ0, fldB1, fldZ1, fldB2, fldZ2);

        fmask initialised = (z[ista] <= z_end) & (z_start < z[ista]);

        fit.SetMask(initialised);
        fit.Extrapolate(z[ista], fld);
        fit.AddMsInMaterial(fParameters.GetMaterialThickness(ista, tr.X(), tr.Y()));
        fit.EnergyLossCorrection(fParameters.GetMaterialThickness(ista, tr.X(), tr.Y()), fvec(-1.f));
        fit.SetMask(initialised && w[ista]);
        fit.FilterXY(cov_xy[ista], x[ista], y[ista]);
        fit.FilterTime(time[ista], dt2[ista], sta[ista].timeInfo);

        fldB2 = fldB1;
        fldZ2 = fldZ1;
        fldB1 = fldB0;
        fldZ1 = fldZ0;
      }

      //fit.SetQp0(tr.Qp());
      //fit.SetMask(fmask::One());
      //fit.MeasureVelocityWithQp();
      //fit.FilterVi(TrackParamV::kClightNsInv);

      TrackParamV Tl = fit.Tr();
      if (kGlobal == fTrackingMode) { Tl.Qp() = fitpv.Tr().Qp(); }

      for (int iVec = 0; iVec < nTracks_SIMD; iVec++) {
        t[iVec]->fParLast.Set(Tl, iVec);
      }

    }  // iter
  }
}
