/* Copyright (C) 2010-2021 Frankfurt Institute for Advanced Studies, Goethe-Universität Frankfurt, Frankfurt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Igor Kulakov [committer], Maksym Zyzak, Valentina Akishina */

#include "L1TripletConstructor.h"

#include "CbmL1.h"
#include "CbmL1MCTrack.h"

#include <algorithm>
#include <iostream>

#include "CaToolsDebugger.h"
#include "CaTrackParam.h"
#include "L1Algo.h"
#include "L1Assert.h"
#include "L1Fit.h"
#include "L1Grid.h"
#include "L1HitArea.h"
#include "L1HitPoint.h"

using cbm::ca::tools::Debugger;

L1TripletConstructor::L1TripletConstructor(L1Algo* algo) { fAlgo = algo; }

void L1TripletConstructor::InitStations(int istal, int istam, int istar)
{
  fIsInitialized = false;

  fIstaL = istal;
  fIstaM = istam;
  fIstaR = istar;

  if (fIstaM >= fAlgo->fParameters.GetNstationsActive()) { return; }

  fStaL = &fAlgo->fParameters.GetStation(fIstaL);
  fStaM = &fAlgo->fParameters.GetStation(fIstaM);
  fStaR = &fAlgo->fParameters.GetStation(fIstaR);

  fHitsL = &(fAlgo->fGridPoints[0]) + fAlgo->fGridHitStartIndex[fIstaL];
  fHitsM = &(fAlgo->fGridPoints[0]) + fAlgo->fGridHitStartIndex[fIstaM];
  fHitsR = &(fAlgo->fGridPoints[0]) + fAlgo->fGridHitStartIndex[fIstaR];

  {  // two stations for approximating the field between the target and the left hit
    int sta1 = fIstaL;
    if (sta1 >= fAlgo->fNfieldStations) { sta1 = fAlgo->fNfieldStations - 1; }
    if (sta1 < 1) { sta1 = 1; }
    int sta0 = sta1 / 2;  // station between fIstaL and the target

    assert(0 <= sta0 && sta0 < sta1 && sta1 < fAlgo->fParameters.GetNstationsActive());

    fFld0Sta[0] = &fAlgo->fParameters.GetStation(sta0);
    fFld0Sta[1] = &fAlgo->fParameters.GetStation(sta1);
  }

  {  // three stations for approximating the field between the left and the right hit

    int sta0 = fIstaL;
    int sta1 = fIstaM;
    int sta2 = fIstaM + 1;
    if (sta2 >= fAlgo->fParameters.GetNstationsActive()) {
      sta2 = fIstaM;
      sta1 = fIstaM - 1;
      if (sta0 >= sta1) { sta0 = sta1 - 1; }
    }

    assert(0 <= sta0 && sta0 < sta1 && sta1 < sta2 && sta2 < fAlgo->fParameters.GetNstationsActive());

    fFld1Sta[0] = &fAlgo->fParameters.GetStation(sta0);
    fFld1Sta[1] = &fAlgo->fParameters.GetStation(sta1);
    fFld1Sta[2] = &fAlgo->fParameters.GetStation(sta2);
  }

  fFit.SetParticleMass(fAlgo->GetDefaultParticleMass());
  if (fAlgo->fpCurrentIteration->GetElectronFlag()) { fFit.SetParticleMass(constants::phys::ElectronMass); }
  fFit.SetMask(fmask::One());

  fIsInitialized = true;
}


void L1TripletConstructor::CreateTripletsForHit(int istal, int istam, int istar, ca::HitIndex_t ihl)
{

  InitStations(istal, istam, istar);

  if (!fIsInitialized) return;

  fIhitL = ihl;

  L1HitPoint& hitl = fHitsL[fIhitL];

  // fit a straight line through the target and the left hit.
  {

    /// Get the field approximation. Add the target to parameters estimation.
    /// Propagaete to the middle station of a triplet.

    fFit.SetQp0(fAlgo->fMaxInvMom);

    ca::FieldValue lB, mB, cB, rB _fvecalignment;
    ca::FieldValue l_B_global, centB_global _fvecalignment;

    // get the magnetic field along the track.
    // (suppose track is straight line with origin in the target)

    fvec xl         = hitl.X();
    fvec yl         = hitl.Y();
    fvec zl         = hitl.Z();
    fvec time       = hitl.T();
    fvec timeEr2    = hitl.dT2();
    const fvec dzli = 1.f / (zl - fAlgo->fTargZ);

    const fvec tx = (xl - fAlgo->fTargX) * dzli;
    const fvec ty = (yl - fAlgo->fTargY) * dzli;

    TrackParamV& T = fFit.Tr();

    T.X()    = xl;
    T.Y()    = yl;
    T.Z()    = zl;
    T.Tx()   = tx;
    T.Ty()   = ty;
    T.Qp()   = fvec(0.);
    T.Time() = time;
    T.Vi()   = fvec(0.);

    fvec txErr2 = fAlgo->fMaxSlopePV * fAlgo->fMaxSlopePV / fvec(9.);
    fvec qpErr2 = fAlgo->fMaxInvMom * fAlgo->fMaxInvMom / fvec(9.);

    T.ResetErrors(1., 1., txErr2, txErr2, qpErr2, (staL().timeInfo ? timeEr2 : 1.e6), 1.e2);
    T.C00() = hitl.dX2();
    T.C10() = hitl.dXY();
    T.C11() = hitl.dY2();

    T.Ndf()     = (fAlgo->fpCurrentIteration->GetPrimaryFlag()) ? fvec(2.) : fvec(0.);
    T.NdfTime() = (staL().timeInfo ? -1 : -2);

    // TODO: iteration parameter: "Starting NDF of track parameters"
    // NDF = number of track parameters (6: x, y, tx, ty, qp, time) - number of measured parameters (3: x, y, time) on station or (2: x, y) on target


    // field made by  the left hit, the target and the station istac in-between.
    // is used for extrapolation to the target and to the middle hit
    ca::FieldRegion fld0;
    {
      ca::FieldValue B0 = fFld0Sta[0]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue B1 = fFld0Sta[1]->fieldSlice.GetFieldValueForLine(T);
      fld0.Set(fAlgo->fTargB, fAlgo->fTargZ, B0, fFld0Sta[0]->fZ, B1, fFld0Sta[1]->fZ);
    }

    {  // field, made by the left hit, the middle station and the right station
      // Will be used for extrapolation to the right hit
      ca::FieldValue B0 = fFld1Sta[0]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue B1 = fFld1Sta[1]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue B2 = fFld1Sta[2]->fieldSlice.GetFieldValueForLine(T);
      fFldL.Set(B0, fFld1Sta[0]->fZ, B1, fFld1Sta[1]->fZ, B2, fFld1Sta[2]->fZ);
    }

    // add the target constraint

    fFit.AddTargetToLine(fAlgo->fTargX, fAlgo->fTargY, fAlgo->fTargZ, fAlgo->TargetXYInfo, fld0);

    fFit.AddMsInMaterial(fAlgo->fParameters.GetMaterialThickness(fIstaL, T.GetX(), T.GetY()));

    // extrapolate to the middle hit

    fFit.ExtrapolateLine(staM().fZ, fFldL);

    fTrL = T;
  }

  FindDoublets();
  FitDoublets();
  FindTriplets();
}


void L1TripletConstructor::FindDoublets()
{
  /// Find the doublets. Reformat data into portions of doublets.

  int iMC = -1;
  if (fAlgo->fParameters.DevIsMatchDoubletsViaMc()) {
    int indL = fAlgo->fGridHitStartIndex[fIstaL] + fIhitL;
    iMC      = fAlgo->GetMcTrackIdForGridHit(indL);
    if (iMC < 0) { return; }
  }

  CollectHits(fTrL, fIstaM, fAlgo->fDoubletChi2Cut, iMC, fHitsM_2, fAlgo->fParameters.GetMaxDoubletsPerSinglet());
}


void L1TripletConstructor::FitDoublets()
{

  // ---- Add the middle hits to parameters estimation ----

  Vector<ca::HitIndex_t> hitsMtmp("L1TripletConstructor::hitsMtmp", fHitsM_2);

  fHitsM_2.clear();
  fTracks_2.clear();
  fTracks_2.reserve(hitsMtmp.size());

  bool isMomentumFitted = (fAlgo->fIsTargetField || (staL().fieldStatus != 0) || (staM().fieldStatus != 0));

  for (unsigned int i2 = 0; i2 < hitsMtmp.size(); i2++) {

    const ca::HitIndex_t imh = hitsMtmp[i2];
    const L1HitPoint& hitm = fHitsM[imh];

    if (hitm.IsSuppressed()) continue;

    TrackParamV& T2 = fFit.Tr();
    T2              = fTrL;
    fFit.SetQp0(fvec(0.f));

    fvec x_2 = hitm.X();
    fvec y_2 = hitm.Y();
    fvec z_2 = hitm.Z();
    fvec t_2 = hitm.T();
    L1XYMeasurementInfo cov_2;
    cov_2.C00  = hitm.dX2();
    cov_2.C10  = hitm.dXY();
    cov_2.C11  = hitm.dY2();
    fvec dt2_2 = hitm.dT2();

    // add the middle hit

    fFit.ExtrapolateLineNoField(z_2);

    fFit.FilterXY(cov_2, x_2, y_2);

    fFit.FilterTime(t_2, dt2_2, staM().timeInfo);

    fFit.SetQp0(isMomentumFitted ? fFit.Tr().GetQp() : fAlgo->fMaxInvMom);

    fFit.AddMsInMaterial(fAlgo->fParameters.GetMaterialThickness(fIstaM, T2.GetX(), T2.Y()));

    fFit.SetQp0(fFit.Tr().Qp());

    // check if there are other hits close to the doublet on the same station
    if (L1Algo::kMcbm != fAlgo->fTrackingMode) {
      // TODO: SG: adjust cuts, put them to parameters

      fscal tx = T2.Tx()[0];
      fscal ty = T2.Ty()[0];
      fscal tt = T2.Vi()[0] * sqrt(1. + tx * tx + ty * ty);  // dt/dl * dl/dz

      for (unsigned int j2 = i2 + 1; j2 < hitsMtmp.size(); j2++) {

        L1HitPoint& hitm1 = fHitsM[hitsMtmp[j2]];

        fscal dz = hitm1.Z() - T2.Z()[0];

        if ((staM().timeInfo) && (T2.NdfTime()[0] > -2)) {
          fscal dt = T2.Time()[0] + tt * dz - hitm.T();
          if (fabs(dt) > sqrt(30. * T2.C55()[0]) + hitm1.RangeT()) { continue; }
        }

        fscal dx = T2.GetX()[0] + tx * dz - hitm1.X();
        if (fabs(dx) > 1.3 * (3.5 * sqrt(T2.C00()[0]) + hitm1.RangeX())) { continue; }

        fscal dy = T2.Y()[0] + ty * dz - hitm1.Y();
        if (fabs(dy) > 1.6 * (3.5 * sqrt(T2.C11()[0]) + hitm1.RangeY())) { continue; }

        if (fAlgo->fParameters.DevIsSuppressOverlapHitsViaMc()) {
          int indL  = fAlgo->fGridHitStartIndex[fIstaL] + fIhitL;
          int indM  = fAlgo->fGridHitStartIndex[fIstaM] + imh;
          int indM1 = fAlgo->fGridHitStartIndex[fIstaM] + hitsMtmp[j2];
          int iMC   = fAlgo->GetMcTrackIdForGridHit(indL);
          if ((iMC != fAlgo->GetMcTrackIdForGridHit(indM)) || (iMC != fAlgo->GetMcTrackIdForGridHit(indM1))) {
            continue;
          }
        }
        hitm1.SetIsSuppresed(1);
      }
    }

    fHitsM_2.push_back(imh);
    fTracks_2.push_back(T2);
  }  // i2
}


void L1TripletConstructor::FindTriplets()
{
  if (fIstaR >= fAlgo->fParameters.GetNstationsActive()) { return; }
  FindRightHit();
  FitTriplets();
  StoreTriplets();
}


void L1TripletConstructor::FindRightHit()
{

  /// Add the middle hits to parameters estimation. Propagate to right station.
  /// Find the triplets(right hit). Reformat data in the portion of triplets.

  fFit.SetQp0(fvec(0.));

  if (fIstaR >= fAlgo->fParameters.GetNstationsActive()) return;

  {
    int maxTriplets = fHitsM_2.size() * fAlgo->fParameters.GetMaxTripletPerDoublets();
    fHitsM_3.clear();
    fHitsR_3.clear();
    fHitsM_3.reserve(maxTriplets);
    fHitsR_3.reserve(maxTriplets);
  }
  // ---- Add the middle hits to parameters estimation. Propagate to right station. ----

  for (unsigned int i2 = 0; i2 < fHitsM_2.size(); i2++) {

    fFit.SetTrack(fTracks_2[i2]);
    TrackParamV& T2 = fFit.Tr();

    // extrapolate to the right hit station

    fFit.Extrapolate(staR().fZ, fFldL);

    // ---- Find the triplets(right hit). Reformat data in the portion of triplets. ----

    if (fAlgo->kSts == fAlgo->fTrackingMode && (T2.C44()[0] < 0)) { continue; }
    if (T2.C00()[0] < 0 || T2.C11()[0] < 0 || T2.C22()[0] < 0 || T2.C33()[0] < 0 || T2.C55()[0] < 0) continue;

    if (fabs(T2.Tx()[0]) > fAlgo->fMaxSlope) continue;
    if (fabs(T2.Ty()[0]) > fAlgo->fMaxSlope) continue;

    int iMC = -1;
    if (fAlgo->fParameters.DevIsMatchTripletsViaMc()) {
      int indL = fAlgo->fGridHitStartIndex[fIstaL] + fIhitL;
      int indM = fAlgo->fGridHitStartIndex[fIstaM] + fHitsM_2[i2];
      iMC      = fAlgo->GetMcTrackIdForGridHit(indL);
      if (iMC < 0 || iMC != fAlgo->GetMcTrackIdForGridHit(indM)) { continue; }
    }

    Vector<ca::HitIndex_t> collectedHits;
    CollectHits(T2, fIstaR, fAlgo->fTripletChi2Cut, iMC, collectedHits, fAlgo->fParameters.GetMaxTripletPerDoublets());

    if (collectedHits.size() >= fAlgo->fParameters.GetMaxTripletPerDoublets()) {
      LOG(debug) << "L1: GetMaxTripletPerDoublets==" << fAlgo->fParameters.GetMaxTripletPerDoublets()
                 << " reached in findTripletsStep0()";
      // reject already created triplets for this doublet
      collectedHits.clear();
    }

    for (unsigned int ih = 0; ih < collectedHits.size(); ih++) {
      ca::HitIndex_t irh     = collectedHits[ih];
      const L1HitPoint& hitr = fHitsR[irh];
      if (hitr.IsSuppressed()) continue;

      // pack the triplet

      fHitsM_3.push_back(fHitsM_2[i2]);
      fHitsR_3.push_back(irh);

    }  // search area
  }    // i2
}


void L1TripletConstructor::FitTriplets()
{
  constexpr bool dumpTriplets = 0;

  constexpr int nIterations = 2;

  Tindex n3 = fHitsM_3.size();
  assert(fHitsM_3.size() == fHitsR_3.size());

  fTracks_3.clear();
  fTracks_3.reset(n3, TrackParamV());

  /// Refit Triplets
  if (dumpTriplets) {
    cbm::ca::tools::Debugger::Instance().AddNtuple(
      "triplets", "ev:iter:i0:x0:y0:z0:i1:x1:y1:z1:i2:x2:y2:z2:mc:sta:p:vx:vy:vz:chi2:ndf:chi2time:ndfTime");
  }

  L1Fit fit;
  fit.SetMask(fmask::One());

  if (fAlgo->fpCurrentIteration->GetElectronFlag()) { fit.SetParticleMass(constants::phys::ElectronMass); }
  else {
    fit.SetParticleMass(fAlgo->GetDefaultParticleMass());
  }
  const int NHits = 3;  // triplets

  // prepare data
  int ista[NHits] = {fIstaL, fIstaM, fIstaR};

  ca::Station sta[3];
  for (int is = 0; is < NHits; ++is) {
    sta[is] = fAlgo->fParameters.GetStation(ista[is]);
  };

  bool isMomentumFitted = ((staL().fieldStatus != 0) || (staM().fieldStatus != 0) || (staR().fieldStatus != 0));

  fvec ndfTrackModel = 4;                     // straight line
  ndfTrackModel += isMomentumFitted ? 1 : 0;  // track with momentum

  for (int i3 = 0; i3 < n3; ++i3) {

    // prepare data
    ca::HitIndex_t ihit[NHits] = {fAlgo->fGridHitIds[fIhitL + fAlgo->fGridHitStartIndex[ista[0]]],
                                  fAlgo->fGridHitIds[fHitsM_3[i3] + fAlgo->fGridHitStartIndex[ista[1]]],
                                  fAlgo->fGridHitIds[fHitsR_3[i3] + fAlgo->fGridHitStartIndex[ista[2]]]};

    if (fAlgo->fParameters.DevIsMatchTripletsViaMc()) {
      int mc1 = fAlgo->GetMcTrackIdForCaHit(ihit[0]);
      int mc2 = fAlgo->GetMcTrackIdForCaHit(ihit[1]);
      int mc3 = fAlgo->GetMcTrackIdForCaHit(ihit[2]);
      if ((mc1 != mc2) || (mc1 != mc3)) { continue; }
    }

    fscal x[NHits], y[NHits], z[NHits], t[NHits];
    fscal dt2[NHits];
    L1XYMeasurementInfo cov[NHits];

    for (int ih = 0; ih < NHits; ++ih) {
      const ca::Hit& hit = fAlgo->fInputData.GetHit(ihit[ih]);
      x[ih]            = hit.x;
      y[ih]            = hit.y;
      z[ih]            = hit.z;
      t[ih]            = hit.t;
      cov[ih].C00      = hit.dx2;
      cov[ih].C10      = hit.dxy;
      cov[ih].C11      = hit.dy2;
      dt2[ih]          = hit.dt2;
    };

    // find the field along the track

    ca::FieldValue B[3] _fvecalignment;
    ca::FieldRegion fld _fvecalignment;
    ca::FieldRegion fldTarget _fvecalignment;

    fvec tx[3] = {(x[1] - x[0]) / (z[1] - z[0]), (x[2] - x[0]) / (z[2] - z[0]), (x[2] - x[1]) / (z[2] - z[1])};
    fvec ty[3] = {(y[1] - y[0]) / (z[1] - z[0]), (y[2] - y[0]) / (z[2] - z[0]), (y[2] - y[1]) / (z[2] - z[1])};
    for (int ih = 0; ih < NHits; ++ih) {
      fvec dz = (sta[ih].fZ - z[ih]);
      B[ih]   = sta[ih].fieldSlice.GetFieldValue(x[ih] + tx[ih] * dz, y[ih] + ty[ih] * dz);
    };

    fld.Set(B[0], sta[0].fZ, B[1], sta[1].fZ, B[2], sta[2].fZ);
    fldTarget.Set(fAlgo->fTargB, fAlgo->fTargZ, B[0], sta[0].fZ, B[1], sta[1].fZ);

    TrackParamV& T = fit.Tr();

    // initial parameters
    {
      fvec dz01 = 1. / (z[1] - z[0]);
      T.Tx()    = (x[1] - x[0]) * dz01;
      T.Ty()    = (y[1] - y[0]) * dz01;
      T.Qp()    = 0.;
      T.Vi()    = 0.;
    }

    // repeat several times in order to improve the precision
    for (int iiter = 0; iiter < nIterations; ++iiter) {
      // fit forward
      {
        fit.SetQp0(T.Qp());
        fit.Qp0()(fit.Qp0() > fAlgo->GetMaxInvMom())  = fAlgo->GetMaxInvMom();
        fit.Qp0()(fit.Qp0() < -fAlgo->GetMaxInvMom()) = -fAlgo->GetMaxInvMom();

        T.Qp()   = 0.;
        T.Vi()   = 0.;
        int ih0  = 0;
        T.X()    = x[ih0];
        T.Y()    = y[ih0];
        T.Z()    = z[ih0];
        T.Time() = t[ih0];

        T.ResetErrors(1., 1., 1., 1., 100., (sta[ih0].timeInfo ? dt2[ih0] : 1.e6), 1.e2);
        T.C00() = cov[ih0].C00;
        T.C10() = cov[ih0].C10;
        T.C11() = cov[ih0].C11;

        T.Ndf()     = -ndfTrackModel + 2;
        T.NdfTime() = sta[ih0].timeInfo ? -1 : -2;

        //  add the target constraint
        fit.AddTargetToLine(fAlgo->fTargX, fAlgo->fTargY, fAlgo->fTargZ, fAlgo->TargetXYInfo, fldTarget);

        for (int ih = 1; ih < NHits; ++ih) {
          fit.Extrapolate(z[ih], fld);
          auto radThick = fAlgo->fParameters.GetMaterialThickness(ista[ih], T.X(), T.Y());
          fit.AddMsInMaterial(radThick);
          fit.EnergyLossCorrection(radThick, fvec(-1.f));
          fit.FilterXY(cov[ih], x[ih], y[ih]);
          fit.FilterTime(t[ih], dt2[ih], sta[ih].timeInfo);
        }
      }

      if (iiter == nIterations - 1) break;

      // fit backward
      {
        fit.SetQp0(T.Qp());
        fit.Qp0()(fit.Qp0() > fAlgo->GetMaxInvMom())  = fAlgo->GetMaxInvMom();
        fit.Qp0()(fit.Qp0() < -fAlgo->GetMaxInvMom()) = -fAlgo->GetMaxInvMom();

        int ih0  = NHits - 1;
        T.X()    = x[ih0];
        T.Y()    = y[ih0];
        T.Z()    = z[ih0];
        T.Time() = t[ih0];
        T.Qp()   = 0.;
        T.Vi()   = 0.;

        T.ResetErrors(1., 1., 1., 1., 100., (sta[ih0].timeInfo ? dt2[ih0] : 1.e6), 1.e2);
        T.C00() = cov[ih0].C00;
        T.C10() = cov[ih0].C10;
        T.C11() = cov[ih0].C11;

        T.Ndf()     = -ndfTrackModel + 2;
        T.NdfTime() = sta[ih0].timeInfo ? -1 : -2;

        for (int ih = NHits - 2; ih >= 0; --ih) {
          fit.Extrapolate(z[ih], fld);
          auto radThick = fAlgo->fParameters.GetMaterialThickness(ista[ih], T.X(), T.Y());
          fit.AddMsInMaterial(radThick);
          fit.EnergyLossCorrection(radThick, fvec(1.f));
          fit.FilterXY(cov[ih], x[ih], y[ih]);
          fit.FilterTime(t[ih], dt2[ih], sta[ih].timeInfo);
        }
      }
    }  // for iiter

    fTracks_3[i3] = T;

    if (dumpTriplets) {
      int ih0 = ihit[0];
      int ih1 = ihit[1];
      int ih2 = ihit[2];
      int mc1 = fAlgo->GetMcTrackIdForCaHit(ih0);
      int mc2 = fAlgo->GetMcTrackIdForCaHit(ih1);
      int mc3 = fAlgo->GetMcTrackIdForCaHit(ih2);

      const ca::Hit& h0 = fAlgo->fInputData.GetHit(ih0);
      const ca::Hit& h1 = fAlgo->fInputData.GetHit(ih1);
      const ca::Hit& h2 = fAlgo->fInputData.GetHit(ih2);

      if ((mc1 >= 0) && (mc1 == mc2) && (mc1 == mc3)) {
        const CbmL1MCTrack& mctr = CbmL1::Instance()->GetMcTracks()[mc1];
        cbm::ca::tools::Debugger::Instance().FillNtuple(
          "triplets", mctr.iEvent, fAlgo->isec, ih0, h0.x, h0.y, h0.z, ih1, h1.x, h1.y, h1.z, ih2, h2.x, h2.y, h2.z,
          mc1, fIstaL, mctr.p, mctr.x, mctr.y, mctr.z, (float) T.GetChiSq()[0], (float) T.Ndf()[0],
          (float) T.ChiSqTime()[0], (float) T.NdfTime()[0]);
      }
    }
  }  //i3

}  // FitTriplets


void L1TripletConstructor::StoreTriplets()
{
  /// Selects good triplets and saves them into fTriplets.
  /// Finds neighbouring triplets at the next station.

  Tindex n3 = fHitsM_3.size();

  bool isMomentumFitted = ((staL().fieldStatus != 0) || (staM().fieldStatus != 0) || (staR().fieldStatus != 0));

  fTriplets.clear();
  fTriplets.reserve(n3);

  for (Tindex i3 = 0; i3 < n3; ++i3) {

    TrackParamV& T3 = fTracks_3[i3];

    fscal chi2 = T3.GetChiSq()[0];  // / T3.NDF[0];

    const ca::HitIndex_t ihitl = fIhitL + fAlgo->fGridHitStartIndex[fIstaL];
    const ca::HitIndex_t ihitm = fHitsM_3[i3] + fAlgo->fGridHitStartIndex[fIstaM];
    const ca::HitIndex_t ihitr = fHitsR_3[i3] + fAlgo->fGridHitStartIndex[fIstaR];

    L1_ASSERT(ihitl < fAlgo->fGridHitStopIndex[fIstaL], ihitl << " < " << fAlgo->fGridHitStopIndex[fIstaL]);
    L1_ASSERT(ihitm < fAlgo->fGridHitStopIndex[fIstaM], ihitm << " < " << fAlgo->fGridHitStopIndex[fIstaM]);
    L1_ASSERT(ihitr < fAlgo->fGridHitStopIndex[fIstaR], ihitr << " < " << fAlgo->fGridHitStopIndex[fIstaR]);

    if (!fAlgo->fpCurrentIteration->GetTrackFromTripletsFlag()) {
      if (chi2 > fAlgo->fTripletFinalChi2Cut) { continue; }
    }

    // assert(std::isfinite(chi2) && chi2 > 0);

    if (!std::isfinite(chi2) || chi2 < 0) { continue; }

    //if (!T3.IsEntryConsistent(true, 0)) { continue; }

    fscal qp  = T3.Qp()[0];
    fscal Cqp = T3.C44()[0];

    // TODO: SG: a magic correction that comes from the legacy code
    // removing it leads to a higher ghost ratio
    Cqp += 0.001;

    fTriplets.emplace_back(ihitl, ihitm, ihitr, fIstaL, fIstaM, fIstaR, 0, 0, 0, chi2, qp, Cqp, T3.Tx()[0], T3.C22()[0],
                           T3.Ty()[0], T3.C33()[0], isMomentumFitted);
  }
}


void L1TripletConstructor::CollectHits(const TrackParamV& Tr, const int iSta, const double chi2Cut, const int iMC,
                                       Vector<ca::HitIndex_t>& collectedHits, int maxNhits)
{
  /// Collect hits on a station

  collectedHits.clear();
  collectedHits.reserve(maxNhits);

  const ca::Station& sta     = fAlgo->fParameters.GetStation(iSta);
  const L1HitPoint* hits   = &(fAlgo->fGridPoints[0]) + fAlgo->fGridHitStartIndex[iSta];
  const ca::HitIndex_t nHits = fAlgo->fGridHitStopIndex[iSta] - fAlgo->fGridHitStartIndex[iSta];

  fFit.SetTrack(Tr);
  TrackParamV& T = fFit.Tr();
  //std::cout << T.chi2[0] << std::endl;
  T.ChiSq() = 0.;
  // if make it bigger the found hits will be rejected later because of the chi2 cut.
  const fvec Pick_m22 = (fvec(chi2Cut) - T.GetChiSq());

  const fscal iz         = 1.f / (T.Z()[0] - fAlgo->fParameters.GetTargetPositionZ()[0]);
  const fscal timeError2 = T.C55()[0];
  const fscal time       = T.Time()[0];

  L1HitAreaTime areaTime(fAlgo->vGridTime[iSta], T.X()[0] * iz, T.Y()[0] * iz,
                         (sqrt(Pick_m22 * T.C00()) + fAlgo->fMaxRangeX[iSta] + fAlgo->fMaxDZ * abs(T.Tx()))[0] * iz,
                         (sqrt(Pick_m22 * T.C11()) + fAlgo->fMaxRangeY[iSta] + fAlgo->fMaxDZ * abs(T.Ty()))[0] * iz,
                         time, sqrt(timeError2) * 5 + fAlgo->fMaxRangeT[iSta]);

  for (ca::HitIndex_t ih = 0; (ih < nHits) && ((int) collectedHits.size() < maxNhits);
       ih++) {  // loop over all station hits

    if (!fAlgo->fParameters.DevIsIgnoreHitSearchAreas()) {  // only loop over the hits in the area
      if (!areaTime.GetNext(ih)) { break; }
    }

    const L1HitPoint& hit = hits[ih];
    if (hit.IsSuppressed()) continue;

    if (iMC >= 0) {  // match via MC
      if (iMC != fAlgo->GetMcTrackIdForGridHit(fAlgo->fGridHitStartIndex[iSta] + ih)) { continue; }
    }

    // check y-boundaries
    //TODO: move hardcoded cuts to parameters
    if ((sta.timeInfo) && (T.NdfTime()[0] > -2)) {
      if (fabs(time - hit.T()) > 1.4 * (3.5 * sqrt(timeError2) + hit.RangeT())) continue;
      if (fabs(time - hit.T()) > 30) continue;
    }

    // - check whether hit belong to the window ( track position +\- errors ) -
    const fscal z = hit.Z();

    fvec y, C11;
    fFit.ExtrapolateYC11Line(z, y, C11);

    fscal dy_est = sqrt(Pick_m22[0] * C11[0]) + hit.RangeY();

    fscal xm = hit.X();
    fscal ym = hit.Y();

    const fscal dY = ym - y[0];

    if (fabs(dY) > dy_est) continue;

    // check x-boundaries
    fvec x, C00;
    fFit.ExtrapolateXC00Line(z, x, C00);

    fscal dx_est = sqrt(Pick_m22[0] * C00[0]) + hit.RangeX();

    const fscal dX = xm - x[0];

    if (fabs(dX) > dx_est) continue;

    // check chi2

    fvec C10;
    fFit.ExtrapolateC10Line(z, C10);

    auto [chi2x, chi2u] = L1Fit::GetChi2XChi2U(hit.X(), hit.Y(), hit.dX2(), hit.dXY(), hit.dY2(), x, y, C00, C10, C11);

    // TODO: adjust the cut, cut on chi2x & chi2u separately
    if (!fAlgo->fpCurrentIteration->GetTrackFromTripletsFlag()) {
      if (chi2x[0] > chi2Cut) continue;
      if (chi2x[0] + chi2u[0] > chi2Cut) continue;
    }

    collectedHits.push_back(ih);

  }  // loop over the hits in the area
}
