/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmCaInputQaSetup.cxx
/// @brief  QA task for tracking detector interfaces (implementation)
/// @since  28.08.2023
/// @author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaSetup.h"

#include "CbmMCDataManager.h"

#include "FairRootManager.h"

#include "TAxis.h"

#include "AlgoFairloggerCompat.h"
#include "L1InitManager.h"

using cbm::algo::ca::Parameters;
using cbm::ca::InputQaSetup;

// ---------------------------------------------------------------------------------------------------------------------
//
InputQaSetup::InputQaSetup(int verbose, bool isMCUsed) : CbmQaTask("CbmCaInputQaSetup", verbose, isMCUsed) {}

// ---------------------------------------------------------------------------------------------------------------------
//
bool InputQaSetup::Check() { return true; }


// ---------------------------------------------------------------------------------------------------------------------
//
void InputQaSetup::CheckInit() const
{
  if (IsMCUsed() && !fpMCEventHeader) { throw std::logic_error("MC event header branch is unavailable"); }
  for (int iD = 0; iD < static_cast<int>(L1DetectorID::kEND); ++iD) {
    if (fvbUseDet[iD]) {
      if (!fvpBrHits[iD]) { throw std::logic_error(Form("Hit branch is unavailable for %s", kDetName[iD])); }
      if (IsMCUsed() && !fvpBrPoints[iD]) {
        throw std::logic_error(Form("MC point branch is unavailable for %s", kDetName[iD]));
      }
    }
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void InputQaSetup::FillHistograms()
{
  if (fvbUseDet[L1DetectorID::kMvd]) { this->FillHistogramsDet<L1DetectorID::kMvd>(); }
  if (fvbUseDet[L1DetectorID::kSts]) { this->FillHistogramsDet<L1DetectorID::kSts>(); }
  if (fvbUseDet[L1DetectorID::kMuch]) { this->FillHistogramsDet<L1DetectorID::kMuch>(); }
  if (fvbUseDet[L1DetectorID::kTrd]) { this->FillHistogramsDet<L1DetectorID::kTrd>(); }
  if (fvbUseDet[L1DetectorID::kTof]) { this->FillHistogramsDet<L1DetectorID::kTof>(); }
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus InputQaSetup::InitCanvases()
{
  // Tracking setup by hits
  {
    auto* canv = MakeQaObject<CbmQaCanvas>("c_setup_hits", "Setup by hits", 1000, 1000);
    canv->Divide(1, 2, 0.000001, 0.000001);
    canv->cd(1);
    gPad->SetLogz();
    gPad->SetBottomMargin(0.000001);
    gPad->SetRightMargin(0.05);
    gPad->SetTitle("");
    auto* ph_hit_xz = dynamic_cast<TH2F*>(fph_hit_xz.back()->Clone("h_hit_xz_clone"));
    ph_hit_xz->SetTitle(";z [cm];x [cm]");
    ph_hit_xz->Draw("col");
    this->PutSetupNameOnPad(0.08, 0.83, 0.50, 0.89);
    canv->cd(2);
    gPad->SetLogz();
    gPad->SetTopMargin(0.000001);
    gPad->SetRightMargin(0.05);
    gPad->SetTitle("");
    auto* ph_hit_yz = dynamic_cast<TH2F*>(fph_hit_yz.back()->Clone("h_hit_yz_clone"));
    ph_hit_yz->SetTitle(";z [cm];y [cm]");
    ph_hit_yz->Draw("col");
    this->PutSetupNameOnPad(0.08, 0.93, 0.50, 0.99);


    auto LoBinEdge = [](TAxis* pAxis, double val) { return pAxis->GetBinLowEdge(pAxis->FindBin(val)); };
    auto UpBinEdge = [](TAxis* pAxis, double val) { return pAxis->GetBinUpEdge(pAxis->FindBin(val)); };

    for (int iDet = 0; iDet < static_cast<int>(fvpDetInterface.size()); ++iDet) {
      if (!fvbUseDet[iDet]) { continue; }
      int nSt = fvpDetInterface[iDet]->GetNtrackingStations();
      for (int iSt = 0; iSt < nSt; ++iSt) {
        int iStActive       = fpParameters->GetStationIndexActive(iSt, static_cast<L1DetectorID>(iDet));
        Color_t boxColor    = (iStActive == -1) ? kGray + 1 : kOrange + 2;
        const char* detName = Form("%s-%d", fvpDetInterface[iDet]->GetDetectorName().c_str(), iSt);
        {
          double zMin = LoBinEdge(fph_hit_xz.back()->GetXaxis(), fvZmin[iDet][iSt]);
          double zMax = UpBinEdge(fph_hit_xz.back()->GetXaxis(), fvZmax[iDet][iSt]);
          double xMin = LoBinEdge(fph_hit_xz.back()->GetYaxis(), fvXmin[iDet][iSt]);
          double xMax = UpBinEdge(fph_hit_xz.back()->GetYaxis(), fvXmax[iDet][iSt]);
          canv->cd(1);
          auto* pBox = new TBox(zMin, xMin, zMax, xMax);
          pBox->SetFillStyle(0);
          pBox->SetLineWidth(2);
          pBox->SetLineColor(boxColor);
          pBox->Draw("same");
          auto* pText = new TText(zMin, xMax, detName);
          pText->SetTextColor(boxColor);
          pText->SetTextSize(0.035);
          pText->SetTextAngle(45);
          pText->Draw("same");
        }
        {
          double zMin = LoBinEdge(fph_hit_yz.back()->GetXaxis(), fvZmin[iDet][iSt]);
          double zMax = UpBinEdge(fph_hit_yz.back()->GetXaxis(), fvZmax[iDet][iSt]);
          double yMin = LoBinEdge(fph_hit_yz.back()->GetYaxis(), fvYmin[iDet][iSt]);
          double yMax = UpBinEdge(fph_hit_yz.back()->GetYaxis(), fvYmax[iDet][iSt]);
          canv->cd(2);
          auto* pBox = new TBox(zMin, yMin, zMax, yMax);
          pBox->SetFillStyle(0);
          pBox->SetLineWidth(2);
          pBox->SetLineColor(boxColor);
          pBox->Draw("same");
          auto* pText = new TText(zMin, yMax, detName);
          pText->SetTextSize(0.035);
          pText->SetTextColor(boxColor);
          pText->SetTextAngle(45);
          pText->Draw("same");
        }
      }
    }
  }
  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus InputQaSetup::InitDataBranches()
try {
  LOG(info) << fName << ": initializing... ";
  // Tracking parameters
  {
    fpParameters = std::make_unique<ca::Parameters>();

    L1InitManager manager;
    manager.ReadParametersObject(fsParametersFilename.c_str());
    manager.SendParameters(*fpParameters);
    LOG(info) << "CA tracking parameters object: " << fsParametersFilename << '\n' << fpParameters->ToString(1);
  }

  auto pFairManager = FairRootManager::Instance();
  assert(pFairManager);

  auto pMcManager = IsMCUsed() ? dynamic_cast<CbmMCDataManager*>(pFairManager->GetObject("MCDataManager")) : nullptr;
  fpMCEventHeader = IsMCUsed() ? pMcManager->GetObject("MCEventHeader.") : nullptr;

  fvpBrPoints.fill(nullptr);
  fvpBrHits.fill(nullptr);

  auto InitDetInterface = [&](CbmTrackingDetectorInterfaceBase* ifs, L1DetectorID detID) {
    if (fvbUseDet[detID]) {
      fvpDetInterface[detID] = ifs;
      int nSt                = ifs->GetNtrackingStations();
      fvXmin[detID].resize(nSt, std::numeric_limits<double>::max());
      fvXmax[detID].resize(nSt, std::numeric_limits<double>::lowest());
      fvYmin[detID].resize(nSt, std::numeric_limits<double>::max());
      fvYmax[detID].resize(nSt, std::numeric_limits<double>::lowest());
      fvZmin[detID].resize(nSt, std::numeric_limits<double>::max());
      fvZmax[detID].resize(nSt, std::numeric_limits<double>::lowest());
    }
  };

  auto InitHitBranch = [&](const char* brName, L1DetectorID detID) {
    if (fvbUseDet[detID]) { fvpBrHits[detID] = dynamic_cast<TClonesArray*>(pFairManager->GetObject(brName)); }
  };

  auto InitPointBranch = [&](const char* brName, L1DetectorID detID) {
    if (IsMCUsed() && fvbUseDet[detID]) { fvpBrPoints[detID] = pMcManager->InitBranch(brName); }
  };

  InitDetInterface(CbmMvdTrackingInterface::Instance(), L1DetectorID::kMvd);
  InitDetInterface(CbmStsTrackingInterface::Instance(), L1DetectorID::kSts);
  InitDetInterface(CbmMuchTrackingInterface::Instance(), L1DetectorID::kMuch);
  InitDetInterface(CbmTrdTrackingInterface::Instance(), L1DetectorID::kTrd);
  InitDetInterface(CbmTofTrackingInterface::Instance(), L1DetectorID::kTof);

  InitHitBranch("MvdHit", L1DetectorID::kMvd);
  InitHitBranch("StsHit", L1DetectorID::kSts);
  InitHitBranch("MuchPixelHit", L1DetectorID::kMuch);
  InitHitBranch("TrdHit", L1DetectorID::kTrd);
  InitHitBranch("TofHit", L1DetectorID::kTof);

  InitPointBranch("MvdPoint", L1DetectorID::kMvd);
  InitPointBranch("StsPoint", L1DetectorID::kSts);
  InitPointBranch("MuchPoint", L1DetectorID::kMuch);
  InitPointBranch("TrdPoint", L1DetectorID::kTrd);
  InitPointBranch("TofPoint", L1DetectorID::kTof);

  this->CheckInit();

  LOG(error) << fName << ": initializing... \033[1;32mDone\033[0m";
  return kSUCCESS;
}
catch (const std::logic_error& e) {
  LOG(error) << fName << ": initializing... \033[1;31mFailed\033[0m\nReason: " << e.what();
  return kFATAL;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus InputQaSetup::InitHistograms()
{
  int nStGeo = fpParameters->GetNstationsGeometry();

  MakeQaDirectory("hit_occupancy");
  for (int iD = 0; iD < static_cast<int>(L1DetectorID::kEND); ++iD) {
    if (fvbUseDet[iD]) { MakeQaDirectory(Form("hit_occupancy/%s", kDetName[iD])); }
  }

  fph_hit_xz.resize(nStGeo + 1);
  fph_hit_yz.resize(nStGeo + 1);

  /// TEMPORARY
  constexpr int kNbinsZ = 300;
  constexpr int kNbinsX = 200;
  constexpr int kNbinsY = 200;
  constexpr float kMinZ = -5.;
  constexpr float kMaxZ = 350.;
  constexpr float kMinX = -100.;
  constexpr float kMaxX = 100.;
  constexpr float kMinY = -100.;
  constexpr float kMaxY = 100.;

  for (int iStGeo = 0; iStGeo <= nStGeo; ++iStGeo) {
    auto [detID, iStLoc] = fpParameters->GetStationIndexLocal(iStGeo);
    TString sF           = "";
    TString sN           = "";
    TString sT           = "";
    TString sNsuff       = (iStGeo == nStGeo) ? "" : Form("_st_%s%d", kDetName[detID], iStLoc);
    TString sTsuff       = (iStGeo == nStGeo) ? "" : Form(" in %s station %d", kDetName[detID], iStLoc);

    // Hit occupancy
    sF                 = (iStGeo == nStGeo) ? "hit_occupancy" : TString(Form("hit_occupancy/%s", kDetName[detID]));
    sN                 = (TString) "hit_xz" + sNsuff;
    sT                 = (TString) "hit occupancy in xz-plane" + sTsuff + ";z_{hit} [cm];x_{hit} [cm]";
    fph_hit_xz[iStGeo] = MakeQaObject<TH2F>(sF + "/" + sN, sT, kNbinsZ, kMinZ, kMaxZ, kNbinsX, kMinX, kMaxX);

    sN                 = (TString) "hit_yz" + sNsuff;
    sT                 = (TString) "hit occupancy in yz-plane" + sTsuff + ";z_{hit} [cm];y_{hit} [cm]";
    fph_hit_yz[iStGeo] = MakeQaObject<TH2F>(sF + "/" + sN, sT, kNbinsZ, kMinZ, kMaxZ, kNbinsY, kMinY, kMaxY);
  }

  return kSUCCESS;
}
