/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */


#include "CbmTaskUnpack.h"

#include "CbmDefs.h"
#include "CbmDigiBranchBase.h"
#include "CbmDigiEvent.h"
#include "CbmDigiManager.h"
#include "CbmDigiTimeslice.h"
#include "CbmSourceTs.h"
#include "CbmTrdParFasp.h"
#include "CbmTrdParModAsic.h"
#include "CbmTrdParModDigi.h"
#include "CbmTrdParSetAsic.h"
#include "CbmTrdParSetDigi.h"
#include "CbmTrdParSpadic.h"

#include "MicrosliceDescriptor.hpp"
#include "System.hpp"

#include <FairParAsciiFileIo.h>
#include <FairParamList.h>
#include <FairRunOnline.h>
#include <Logger.h>

#include <TStopwatch.h>

#include <algorithm>
#include <cassert>
#include <cstdint>
#include <iomanip>
#include <memory>
#include <sstream>
#include <vector>

using namespace std;

// -----   Constructor   -----------------------------------------------------
CbmTaskUnpack::CbmTaskUnpack() : FairTask("Unpack") {}
// ---------------------------------------------------------------------------


// -----   Destructor   ------------------------------------------------------
CbmTaskUnpack::~CbmTaskUnpack()
{
  if (fTimeslice) delete fTimeslice;
}
// ---------------------------------------------------------------------------


// -----   Execution   -------------------------------------------------------
void CbmTaskUnpack::Exec(Option_t*)
{

  // --- Reset output branch (CbmDigiTimeslice)
  fTimeslice->Clear();

  // --- Get FLES timeslice
  assert(fSource);
  fles::Timeslice* timeslice = fSource->GetTimeslice();
  assert(timeslice);

  // --- Timer and counters
  TStopwatch timer;
  timer.Start();

  // --- Unpack the timeslice
  auto result = fUnpack(timeslice);
  fTimeslice->fData = result.first.ToStorable();

  // --- Timeslice log
  timer.Stop();
  stringstream logOut;
  logOut << setw(15) << left << GetName() << " [";
  logOut << fixed << setw(8) << setprecision(1) << right << timer.RealTime() * 1000. << " ms] ";
  logOut << "TS " << fNumTs << " (index " << timeslice->index() << ")";
  logOut << ", components " << result.second.fNumCompUsed << " / " << timeslice->num_components();
  logOut << ", microslices " << result.second.fNumMs;
  logOut << ", input rate " << double(result.second.fNumBytes) / timer.RealTime() / 1.e6 << " MB/s";
  logOut << ", digis " << result.second.fNumDigis;
  LOG(info) << logOut.str();

#if !defined(__CLING__) && !defined(__ROOTCLING__)
  if (fMonitor) {
    fMonitor->QueueMetric(GetName(), {{"host", fHostname}},
                          {{"realtime", timer.RealTime()},
                           {"cputime", timer.CpuTime()},
                           {"input_size", result.second.fNumBytes},
                           {"input_rate", double(result.second.fNumBytes) / timer.RealTime()},
                           {"digis", result.second.fNumDigis}});
  }
#endif

  // --- Run statistics
  fNumTs++;
  fNumMs += result.second.fNumMs;
  fNumBytes += result.second.fNumBytes;
  fNumDigis += result.second.fNumDigis;
  fTime += timer.RealTime();
}
// ----------------------------------------------------------------------------


// -----   End-of-run action   ------------------------------------------------
void CbmTaskUnpack::Finish()
{
  double timePerTs = 1000. * fTime / double(fNumTs);  // in ms
  double rate      = fNumBytes / 1.e6 / fTime;        // in MB/s
  LOG(info) << "=====================================";
  LOG(info) << GetName() << ": Run summary";
  LOG(info) << "Timeslices     : " << fNumTs;
  LOG(info) << "Microslices    : " << fNumMs;
  LOG(info) << "Digis          : " << fNumDigis;
  LOG(info) << "Av. input rate : " << fixed << setprecision(2) << rate << " MB/s";
  LOG(info) << "Time / TS      : " << fixed << setprecision(2) << timePerTs << " ms";
  LOG(info) << "=====================================";
}
// ----------------------------------------------------------------------------


// -----   Initialisation   ---------------------------------------------------
InitStatus CbmTaskUnpack::Init()
{
  LOG(info) << "==================================================";
  LOG(info) << GetName() << ": Initialising...";

  // --- Get hostname
  fHostname = fles::system::current_hostname();

  // --- Get source instance
  fSource = dynamic_cast<CbmSourceTs*>(FairRunOnline::Instance()->GetSource());
  if (fSource == nullptr) {
    LOG(error) << GetName() << ": No valid source class registered!";
    return kFATAL;
  }
  LOG(info) << "--- Found CbmSourceTs instance";

  // --- Get FairRootManager instance
  FairRootManager* ioman = FairRootManager::Instance();
  assert(ioman);

  // --- Register output array (CbmDigiTimeslice)
  if (ioman->GetObject("DigiTimeslice")) {
    LOG(fatal) << GetName() << ": Branch DigiTimeslice already exists!";
    return kFATAL;
  }
  fTimeslice = new CbmDigiTimeslice();
  ioman->RegisterAny("DigiTimeslice.", fTimeslice, IsOutputBranchPersistent("DigiTimeslice."));
  LOG(info) << "--- Registered branch DigiTimeslice.";

  // ---- Initialize unpacker
  fUnpack.fTrdConfig   = InitTrdReadoutConfig();
  fUnpack.fTrd2dConfig = InitTrd2dReadoutConfig();
  fUnpack.Init();

  return kSUCCESS;
}
// ----------------------------------------------------------------------------

// -----   Initialisation   ---------------------------------------------------
cbm::algo::trd2d::ReadoutConfig CbmTaskUnpack::InitTrd2dReadoutConfig()
{
  // Output object
  cbm::algo::trd2d::ReadoutConfig Trd2dConfig;

  // Initialize input files
  FairParAsciiFileIo asciiInput;
  std::string digiparfile = Form("%s/parameters/trd/trd_v22h_mcbm.digi.par", std::getenv("VMCWORKDIR"));
  std::string asicparfile = Form("%s/parameters/trd/trd_v22h_mcbm.asic.par", std::getenv("VMCWORKDIR"));

  // Read the .digi file and store result
  CbmTrdParSetDigi digiparset;
  if (asciiInput.open(digiparfile.data())) { digiparset.init(&asciiInput); }
  asciiInput.close();

  // Read the .asic file and store result
  CbmTrdParSetAsic asicparset;
  if (asciiInput.open(asicparfile.data())) { asicparset.init(&asciiInput); }
  asciiInput.close();

  // Map (moduleId) -> (array of crobId)
  std::map<uint32_t, uint16_t[NCROBMOD]> crobMap;
  // Map (equipId, asicId, chanId) -> (pad address, mask flag, daq offset [FASP clk])
  std::map<size_t, std::map<size_t, std::map<size_t, std::tuple<int32_t, bool, uint64_t>>>> channelMap;

  // Loop through a list of module IDs from the .digi file (can in principle contradict crob_map).
  for (auto entry : digiparset.GetModuleMap()) {

    const auto moduleId = entry.first;

    // Get ASIC parameters for this module
    CbmTrdParModAsic* setDet = static_cast<CbmTrdParModAsic*>(asicparset.GetModulePar(moduleId));
    if (!setDet) continue;
    if (setDet->GetAsicType() != CbmTrdDigi::eCbmTrdAsicType::kFASP) continue;
    auto digipar = entry.second;

    const int* crobs = setDet->GetCrobAddresses();
    for (int icrob(0); icrob < NCROBMOD; icrob++)
      crobMap[moduleId][icrob] = crobs[icrob];

    // Loop through ASICs for this module
    std::vector<int32_t> addresses;
    setDet->GetAsicAddresses(&addresses);
    for (auto add : addresses) {
      //Get local IDs for this component / equipment.
      const int32_t fasp_in_mod = add - 1000 * moduleId;
      const int32_t fasp_in_eq  = fasp_in_mod % (NFASPCROB);
      const int32_t crob_in_mod = fasp_in_mod / (NFASPCROB);
      const uint16_t eq_id      = crobMap[moduleId][crob_in_mod];

      // ASIC parameter set
      CbmTrdParFasp* fasppar = (CbmTrdParFasp*) setDet->GetAsicPar(add);

      // Loop through channels for this ASIC and fill map
      for (int chan = 0; chan < fasppar->GetNchannels(); chan++) {
        const CbmTrdParFaspChannel* faspch = fasppar->GetChannel(chan);
        const int32_t pad                  = fasppar->GetPadAddress(chan) * (faspch->HasPairingR() ? 1 : -1);
        const bool isMasked                = faspch->IsMasked();
        uint64_t daq_offset                = 0;
        if (((CbmTrdParModDigi*) digipar)->GetPadRow(pad) % 2 == 0) daq_offset = 3;
        channelMap[eq_id][fasp_in_eq][chan] = std::make_tuple(pad, isMasked, daq_offset);
      }
    }
  }

  Trd2dConfig.InitComponentMap(crobMap);
  Trd2dConfig.InitChannelMap(channelMap);
  return Trd2dConfig;
}

// -----   Initialisation   ---------------------------------------------------
cbm::algo::trd::ReadoutConfig CbmTaskUnpack::InitTrdReadoutConfig()
{
  // Output object
  cbm::algo::trd::ReadoutConfig TrdConfig;

  std::string trdparfile = Form("%s/parameters/trd/trd_v22h_mcbm.asic.par", std::getenv("VMCWORKDIR"));

  CbmTrdParSetAsic trdpar;

  FairParAsciiFileIo asciiInput;
  if (asciiInput.open(trdparfile.data())) { trdpar.init(&asciiInput); }

  FairParamList parlist;
  trdpar.putParams(&parlist);

  std::vector<int> moduleIds(trdpar.GetNrOfModules());
  parlist.fill("ModuleId", moduleIds.data(), moduleIds.size());

  std::map<size_t, std::map<size_t, std::map<size_t, size_t>>> addressMap;  //[criId][crobId][elinkId] -> asicAddress
  std::map<size_t, std::map<size_t, std::map<size_t, std::map<size_t, size_t>>>>
    channelMap;  //[criId][crobId][elinkId][chanId] -> chanAddress

  for (auto module : moduleIds) {
    CbmTrdParModAsic* moduleSet = (CbmTrdParModAsic*) trdpar.GetModulePar(module);

    // Skip entries for "Fasp" modules in .asic.par file
    if (moduleSet->GetAsicType() != CbmTrdDigi::eCbmTrdAsicType::kSPADIC) continue;

    std::vector<int> asicAddresses;
    moduleSet->GetAsicAddresses(&asicAddresses);

    for (auto address : asicAddresses) {
      CbmTrdParSpadic* asicPar = (CbmTrdParSpadic*) moduleSet->GetAsicPar(address);
      const uint16_t criId     = asicPar->GetCriId();
      const uint8_t crobId     = asicPar->GetCrobId();
      const uint8_t elinkId    = asicPar->GetElinkId(0);
      if (elinkId >= 98) { continue; }  // Don't add not connected asics to the map
      addressMap[criId][crobId][elinkId]     = address;
      addressMap[criId][crobId][elinkId + 1] = address;

      const uint8_t numChans = 16;
      for (uint8_t chan = 0; chan < numChans; chan++) {
        auto asicChannelId                       = (elinkId % 2) == 0 ? chan : chan + numChans;
        auto chanAddr                            = asicPar->GetChannelAddresses().at(asicChannelId);
        channelMap[criId][crobId][elinkId][chan] = chanAddr;
      }
      for (uint8_t chan = 0; chan < numChans; chan++) {
        auto asicChannelId                           = (elinkId + 1 % 2) == 0 ? chan : chan + numChans;
        auto chanAddr                                = asicPar->GetChannelAddresses().at(asicChannelId);
        channelMap[criId][crobId][elinkId + 1][chan] = chanAddr;
      }
      LOG(debug) << "componentID " << asicPar->GetComponentId() << " "
                 << "address " << address << " key " << criId << " " << unsigned(crobId) << " " << unsigned(elinkId);
    }
  }
  TrdConfig.Init(addressMap, channelMap);
  LOG(debug) << TrdConfig.PrintReadoutMap();
  return TrdConfig;
}

ClassImp(CbmTaskUnpack)
